from decimal import Decimal

from django.contrib.messages import error, info
from django.db.models import (
    Case,
    F,
    Sum,
    Value,
    When,
)
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from wbcore import viewsets
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.strings import format_number
from wbcore.viewsets.mixins import OrderableMixin

from wbportfolio.import_export.resources.trades import OrderProposalTradeResource
from wbportfolio.models import Order, OrderProposal
from wbportfolio.serializers import (
    OrderOrderProposalListModelSerializer,
    OrderOrderProposalModelSerializer,
    ReadOnlyOrderOrderProposalModelSerializer,
)

from ...filters.orders import OrderFilterSet
from ..mixins import UserPortfolioRequestPermissionMixin
from .configs import (
    OrderOrderProposalButtonConfig,
    OrderOrderProposalDisplayConfig,
    OrderOrderProposalEndpointConfig,
)


class OrderOrderProposalModelViewSet(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, OrderableMixin, viewsets.ModelViewSet
):
    IMPORT_ALLOWED = True
    ordering = (
        "order_proposal",
        "order",
    )
    ordering_fields = (
        "underlying_instrument__name",
        "underlying_instrument_isin",
        "underlying_instrument_ticker",
        "underlying_instrument_refinitiv_identifier_code",
        "underlying_instrument_instrument_type",
        "target_weight",
        "effective_weight",
        "effective_shares",
        "target_shares",
        "shares",
        "weighting",
    )
    IDENTIFIER = "wbportfolio:orderproposal"
    search_fields = ("underlying_instrument__name",)
    queryset = Order.objects.none()
    filterset_class = OrderFilterSet

    display_config_class = OrderOrderProposalDisplayConfig
    endpoint_config_class = OrderOrderProposalEndpointConfig
    serializer_class = OrderOrderProposalModelSerializer
    button_config_class = OrderOrderProposalButtonConfig

    @cached_property
    def order_proposal(self):
        return get_object_or_404(OrderProposal, pk=self.kwargs["order_proposal_id"])

    @cached_property
    def portfolio_total_asset_value(self):
        return self.order_proposal.portfolio_total_asset_value

    def has_import_permission(self, request) -> bool:  # allow import only on draft order proposal
        return super().has_import_permission(request) and self.order_proposal.status == OrderProposal.Status.DRAFT

    def get_import_resource_kwargs(self):
        resource_kwargs = super().get_import_resource_kwargs()
        resource_kwargs["columns_mapping"] = {"underlying_instrument": "underlying_instrument__isin"}
        return resource_kwargs

    def get_resource_class(self):
        return OrderProposalTradeResource

    def get_aggregates(self, queryset, *args, **kwargs):
        agg = {}
        if queryset.exists():
            noncash_aggregates = queryset.filter(underlying_instrument__is_cash=False).aggregate(
                sum_target_weight=Sum(F("target_weight")),
                sum_effective_weight=Sum(F("effective_weight")),
                sum_target_total_value_fx_portfolio=Sum(F("target_total_value_fx_portfolio")),
                sum_effective_total_value_fx_portfolio=Sum(F("effective_total_value_fx_portfolio")),
            )
            # weights aggregates
            cash_sum_effective_weight = (
                self.order_proposal.total_effective_portfolio_weight - noncash_aggregates["sum_effective_weight"]
            )
            cash_sum_target_cash_weight = Decimal("1.0") - noncash_aggregates["sum_target_weight"]
            noncash_sum_effective_weight = noncash_aggregates["sum_effective_weight"] or Decimal(0)
            noncash_sum_target_weight = noncash_aggregates["sum_target_weight"] or Decimal(0)
            sum_buy_weight = queryset.filter(weighting__gte=0).aggregate(s=Sum(F("weighting")))["s"] or Decimal(0)
            sum_sell_weight = queryset.filter(weighting__lt=0).aggregate(s=Sum(F("weighting")))["s"] or Decimal(0)

            # shares aggregates
            cash_sum_effective_total_value_fx_portfolio = cash_sum_effective_weight * self.portfolio_total_asset_value
            cash_sum_target_total_value_fx_portfolio = cash_sum_target_cash_weight * self.portfolio_total_asset_value
            noncash_sum_effective_total_value_fx_portfolio = noncash_aggregates[
                "sum_effective_total_value_fx_portfolio"
            ] or Decimal(0)
            noncash_sum_target_total_value_fx_portfolio = noncash_aggregates[
                "sum_target_total_value_fx_portfolio"
            ] or Decimal(0)
            sum_buy_total_value_fx_portfolio = queryset.filter(total_value_fx_portfolio__gte=0).aggregate(
                s=Sum(F("total_value_fx_portfolio"))
            )["s"] or Decimal(0)
            sum_sell_total_value_fx_portfolio = queryset.filter(total_value_fx_portfolio__lt=0).aggregate(
                s=Sum(F("total_value_fx_portfolio"))
            )["s"] or Decimal(0)

            agg = {
                "effective_weight": {
                    "Cash": format_number(cash_sum_effective_weight, decimal=Order.ORDER_WEIGHTING_PRECISION),
                    "Non-Cash": format_number(noncash_sum_effective_weight, decimal=Order.ORDER_WEIGHTING_PRECISION),
                    "Total": format_number(
                        noncash_sum_effective_weight + cash_sum_effective_weight,
                        decimal=Order.ORDER_WEIGHTING_PRECISION,
                    ),
                },
                "target_weight": {
                    "Cash": format_number(cash_sum_target_cash_weight, decimal=Order.ORDER_WEIGHTING_PRECISION),
                    "Non-Cash": format_number(noncash_sum_target_weight, decimal=Order.ORDER_WEIGHTING_PRECISION),
                    "Total": format_number(
                        cash_sum_target_cash_weight + noncash_sum_target_weight,
                        decimal=Order.ORDER_WEIGHTING_PRECISION,
                    ),
                },
                "effective_total_value_fx_portfolio": {
                    "Cash": format_number(cash_sum_effective_total_value_fx_portfolio, decimal=6),
                    "Non-Cash": format_number(noncash_sum_effective_total_value_fx_portfolio, decimal=6),
                    "Total": format_number(
                        cash_sum_effective_total_value_fx_portfolio + noncash_sum_effective_total_value_fx_portfolio,
                        decimal=6,
                    ),
                },
                "target_total_value_fx_portfolio": {
                    "Cash": format_number(cash_sum_target_total_value_fx_portfolio, decimal=6),
                    "Non-Cash": format_number(noncash_sum_target_total_value_fx_portfolio, decimal=6),
                    "Total": format_number(
                        cash_sum_target_total_value_fx_portfolio + noncash_sum_target_total_value_fx_portfolio,
                        decimal=6,
                    ),
                },
                "weighting": {
                    "Cash Flow": format_number(
                        sum_sell_weight + sum_buy_weight,
                        decimal=Order.ORDER_WEIGHTING_PRECISION,
                    ),
                    "Buy": format_number(sum_buy_weight, decimal=Order.ORDER_WEIGHTING_PRECISION),
                    "Sell": format_number(sum_sell_weight, decimal=Order.ORDER_WEIGHTING_PRECISION),
                },
                "total_value_fx_portfolio": {
                    "Cash Flow": format_number(
                        cash_sum_target_total_value_fx_portfolio - cash_sum_effective_total_value_fx_portfolio,
                        decimal=6,
                    ),
                    "Buy": format_number(sum_buy_total_value_fx_portfolio, decimal=6),
                    "Sell": format_number(sum_sell_total_value_fx_portfolio, decimal=6),
                },
            }

        return agg

    def get_serializer_class(self):
        if self.order_proposal.status != OrderProposal.Status.DRAFT:
            return ReadOnlyOrderOrderProposalModelSerializer
        if not self.new_mode and "pk" not in self.kwargs:
            serializer_base_class = OrderOrderProposalListModelSerializer
        else:
            serializer_base_class = OrderOrderProposalModelSerializer
        if not self.order_proposal.portfolio_total_asset_value:

            class OnlyWeightSerializerClass(serializer_base_class):
                class Meta(serializer_base_class.Meta):
                    read_only_fields = (
                        "order_type",
                        "effective_shares",
                        "effective_total_value_fx_portfolio",
                        "has_warnings",
                        "shares",
                        "target_shares",
                        "total_value_fx_portfolio",
                        "target_total_value_fx_portfolio",
                    )

            return OnlyWeightSerializerClass
        return serializer_base_class

    def add_messages(self, request, queryset=None, paginated_queryset=None, instance=None, initial=False):
        if self.orders.exists() and self.order_proposal.status in [
            OrderProposal.Status.PENDING,
            OrderProposal.Status.DRAFT,
        ]:
            total_target_weight = self.orders.aggregate(c=Sum(F("target_weight")))["c"] or Decimal(0)
            if round(total_target_weight, 8) != 1:
                info(
                    request,
                    "The total target weight does not equal 1. To avoid automatic cash allocation, please adjust the order weights to sum up to 1. Otherwise, a cash component will be added when this order proposal is submitted.",
                )
            if self.orders.filter(has_warnings=True).exists():
                error(
                    request,
                    "Some orders failed preparation. To resolve this, please revert the order proposal to draft, review and correct the orders, and then resubmit.",
                )

    @cached_property
    def orders(self):
        qs = self.order_proposal.get_orders()
        if not self.is_portfolio_manager:
            return qs.none()
        return qs

    def get_queryset(self):
        return self.orders.annotate(  # .exclude(underlying_instrument__is_cash=True)
            underlying_instrument_isin=F("underlying_instrument__isin"),
            underlying_instrument_ticker=F("underlying_instrument__ticker"),
            underlying_instrument_refinitiv_identifier_code=F("underlying_instrument__refinitiv_identifier_code"),
            underlying_instrument_instrument_type=Case(
                When(
                    underlying_instrument__parent__is_security=True,
                    then=F("underlying_instrument__parent__instrument_type__short_name"),
                ),
                default=F("underlying_instrument__instrument_type__short_name"),
            ),
            underlying_instrument_exchange=F("underlying_instrument__exchange__name"),
            effective_total_value_fx_portfolio=F("effective_weight") * Value(self.portfolio_total_asset_value),
            target_total_value_fx_portfolio=F("target_weight") * Value(self.portfolio_total_asset_value),
            portfolio_currency=F("portfolio__currency__symbol"),
            security=F("underlying_instrument__parent"),
            company=F("underlying_instrument__parent__parent"),
        ).select_related(
            "underlying_instrument", "underlying_instrument__parent", "underlying_instrument__parent__parent"
        )



from typing import Optional

from django.utils.translation import gettext_lazy as _
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.formatting import Condition, Operator
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbfdm.contrib.metric.viewsets.configs.utils import (
    PERFORMANCE_FORMATTING,
    get_performance_fields,
)


class ProductDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="information",
                    label=_("Information"),
                    open_by_default=False,
                    children=[
                        dp.Field(key="name_repr", label="Name", width=250),
                        dp.Field(key="parent", label="Parent"),
                        dp.Field(key="isin", label="ISIN"),
                        dp.Field(key="ticker", label="Ticker", show="open"),
                        dp.Field(key="bank", label="Bank"),
                        dp.Field(key="description", label="Description", show="open"),
                        dp.Field(key="currency", label="Currency", width=100, show="open"),
                        dp.Field(key="classifications", label="Classifications", show="open"),
                        dp.Field(key="white_label_customers", label="Customers", width=100, show="open"),
                    ],
                ),
                dp.Field(
                    label="Fees",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="current_bank_fees", label="Bank Fees", width=100, show="open"),
                        dp.Field(key="current_management_fees", label="Management Fees", width=100, show="open"),
                        dp.Field(key="current_total_issuer_fees", label="Total Issuer Fees", width=100),
                        dp.Field(key="current_performance_fees", label="Performance Fees", width=100, show="open"),
                    ],
                ),
                dp.Field(
                    label="Last Valuation",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="last_valuation_date", show="open", label="Date", width=150),
                        dp.Field(key="net_value", label="Valuation", width=150, tooltip=dp.Tooltip(key="market_data")),
                    ],
                ),
                dp.Field(
                    key=None,
                    label=_("Performance"),
                    open_by_default=False,
                    children=[
                        dp.Field(key="performance_is_estimated", label="Estimated", width=50, show="open"),
                        dp.Field(key="performance_date", label="Date", width=85),
                        *get_performance_fields(with_comparison_performances=True),
                        dp.Field(key="assets_under_management", label="AUM", width=100, show="open"),
                        dp.Field(key="assets_under_management_usd", label="AUM ($)", width=100),
                    ],
                ),
                dp.Field(
                    label="NNM",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(
                            key="nnm_weekly",
                            label="Weekly",
                            width=150,
                            show="open",
                            formatting_rules=PERFORMANCE_FORMATTING,
                        ),
                        dp.Field(
                            key="nnm_monthly",
                            label="Monthly",
                            width=150,
                            show="open",
                            formatting_rules=PERFORMANCE_FORMATTING,
                        ),
                        dp.Field(
                            key="nnm_year_to_date", label="YTD", width=150, formatting_rules=PERFORMANCE_FORMATTING
                        ),
                        dp.Field(
                            key="nnm_yearly",
                            label="Yearly",
                            width=150,
                            show="open",
                            formatting_rules=PERFORMANCE_FORMATTING,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="is_invested",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": "#ADD8E6"},
                            condition=Condition(Operator("=="), False),
                        ),
                    ],
                )
            ],
            legends=[
                dp.Legend(
                    key="is_invested",
                    items=[dp.LegendItem(icon="#ADD8E6", label="Internal (Double Accounting)", value=False)],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "name"), repeat_field(2, "name_repr")],
                ["inception_date", "delisted_date", "isin", "id_repr"],
                ["share_price", "issue_price", "initial_high_water_mark", "currency"],
                [repeat_field(2, "bank"), repeat_field(2, "parent")],
                [repeat_field(4, "tags")],
                [repeat_field(4, "description")],
                [repeat_field(4, "classifications_section")],
                [repeat_field(4, "fees_section")],
                [repeat_field(4, "information_section")],
                [repeat_field(4, "white_label_customers_section")],
                [repeat_field(4, "portfolios_section")],
                [repeat_field(4, "related_instruments_section")],
                [repeat_field(4, "baseindexpositions_section")],
                [repeat_field(4, "portfoliorole_section")],
                [repeat_field(4, "preferredclassification_section")],
            ],
            [
                create_simple_section(
                    "classifications_section", _("Classifications"), [["classifications"]], "classifications"
                ),
                create_simple_section(
                    "fees_section",
                    _("Fees"),
                    [
                        ["current_bank_fees", "current_management_fees", "current_performance_fees", "."],
                    ],
                    "fees",
                ),
                create_simple_section(
                    "information_section",
                    _("Information"),
                    [
                        [repeat_field(2, "asset_class"), "type_of_return", "legal_structure"],
                        ["jurisdiction", "investment_index", repeat_field(2, "risk_scale")],
                        [repeat_field(2, "external_webpage"), repeat_field(2, "liquidity")],
                    ],
                    "information",
                ),
                create_simple_section(
                    "white_label_customers_section",
                    _("White Label Customers"),
                    [["white_label_customers"]],
                    "white_label_customers",
                ),
                create_simple_section("portfolios_section", _("Portfolios"), [["portfolios"]], "portfolios"),
                create_simple_section(
                    "related_instruments_section",
                    _("Related Instruments"),
                    [["related_instruments"]],
                    "related_instruments",
                ),
                create_simple_section(
                    "baseindexpositions_section", _("Index"), [["baseindexpositions"]], "baseindexpositions"
                ),
                create_simple_section("portfoliorole_section", _("Roles"), [["portfoliorole"]], "portfoliorole"),
                create_simple_section(
                    "preferredclassification_section",
                    _("Preferred Classification per Instrument"),
                    [["preferredclassification"]],
                    "preferredclassification",
                ),
            ],
        )


class ProductCustomerDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="isin", label="ISIN"),
                dp.Field(key="ticker", label="Ticker"),
                dp.Field(key="net_value", label="Net Value"),
                dp.Field(key="bank_repr", label="Bank"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(3, "name")],
                ["isin", "ticker", "bank"],
                ["share_price", "issue_price", "currency"],
                [repeat_field(3, "description")],
            ]
        )


class ProductPerformanceFeesDisplayConfig(ProductDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="computed_str", label="Title", width=Unit.PIXEL(400)),
                dp.Field(key="isin", label="ISIN", width=Unit.PIXEL(150)),
                dp.Field(key="sum_management_fees", label="Mgmt", width=Unit.PIXEL(150)),
                dp.Field(key="sum_management_fees_usd", label="Mgmt ($)", width=Unit.PIXEL(150)),
                dp.Field(key="sum_performance_fees_net", label="Perf", width=Unit.PIXEL(150)),
                dp.Field(key="sum_performance_fees_net_usd", label="Perf ($)", width=Unit.PIXEL(150)),
                dp.Field(key="sum_total", label="Total", width=Unit.PIXEL(150)),
                dp.Field(
                    key="sum_total_usd",
                    label="$ Total",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"fontWeight": "bold"},
                        ),
                    ],
                    width=Unit.PIXEL(150),
                ),
            ]
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class TradeEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:trade-list", args=[], request=self.request)

    # def get_instance_endpoint(self, **kwargs):
    #     try:
    #         if self.request.user.has_perm("wbportfolio.administrate_trade"):
    #             obj = self.view.get_object()
    #             return reverse("wbportfolio:trade-detail", args=[obj.id], request=self.request)
    #     except AssertionError:
    #         pass
    #     return reverse("wbportfolio:trade-list", request=self.request)

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None


class TradeInstrumentEndpointConfig(TradeEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-trade-list", args=[self.view.kwargs["instrument_id"]], request=self.request
        )


class TradePortfolioEndpointConfig(TradeEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-trade-list", args=[self.view.kwargs["portfolio_id"]], request=self.request
        )


class CustodianDistributionInstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-custodiandistribution-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class SubscriptionRedemptionEndpointConfig(TradeEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:subscriptionredemption-list", args=[], request=self.request)



# Generated by Django 5.0.9 on 2024-10-18 07:35

from datetime import date, timedelta

import django.contrib.postgres.fields.ranges
from django.db import migrations
from psycopg.types.range import DateRange


def migrate_is_invested(apps, schema_editor):
    Portfolio = apps.get_model("wbportfolio", "Portfolio")
    from wbportfolio.models import Product

    for portfolio in Portfolio.objects.filter(is_invested=True):
        if portfolio.instruments.exists():
            min_inception_date = portfolio.instruments.earliest("inception_date").inception_date
        else:
            min_inception_date = date.min
        portfolio.invested_timespan = DateRange(min_inception_date if min_inception_date else date.min, date.max)
        portfolio.save()
    for product in Product.objects.filter(
        portfolios__in=Portfolio.objects.filter(is_invested=False).values("id")
    ).distinct():
        portfolio = product.portfolio
        if product.delisted_date:
            portfolio.invested_timespan = DateRange(product.inception_date, product.delisted_date)
        else:
            portfolio.invested_timespan = DateRange(
                product.inception_date if product.inception_date else date.min,
                product.primary_portfolio.assets.filter(underlying_instrument__instrument_type__key="product")
                .earliest("date")
                .date
                - timedelta(days=1),
            )
        portfolio.save()


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0068_trade_internal_trade_trade_marked_as_internal_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="portfolio",
            name="invested_timespan",
            field=django.contrib.postgres.fields.ranges.DateRangeField(
                blank=True, help_text="Define when this portfolio is considered invested", null=True
            ),
        ),
        migrations.RunPython(migrate_is_invested),
        migrations.RemoveField(
            model_name="portfolio",
            name="is_invested",
        ),
    ]



from typing import TYPE_CHECKING, Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbfdm.contrib.metric.viewsets.configs.display import (
    InstrumentMetricPivotedListDisplayConfig,
)

if TYPE_CHECKING:
    from wbportfolio.viewsets.assets import CompositionModelPortfolioPandasView


class AssetPositionDisplayConfig(InstrumentMetricPivotedListDisplayConfig):
    LEGENDS = [
        dp.Legend(
            items=[
                dp.LegendItem(icon=WBColor.YELLOW.value, label="Estimated"),
                dp.LegendItem(icon=WBColor.GREEN_LIGHT.value, label="Real"),
            ],
        ),
    ]
    FORMATTING = [
        dp.Formatting(
            column="is_estimated",
            formatting_rules=[
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.YELLOW.value},
                    condition=("==", True),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                    condition=("==", False),
                ),
            ],
        ),
    ]

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                # dp.Field(key="date", label="Date"),
                # dp.Field(key="is_estimated", label="Temporary", width=Unit.PIXEL(100)),
                dp.Field(key="underlying_instrument", label="Name"),
                dp.Field(key="underlying_instrument_isin", label="ISIN"),
                dp.Field(key="underlying_instrument_ticker", label="Ticker"),
                dp.Field(key="portfolio", label="Portfolio"),
                dp.Field(key="exchange", label="Exchange"),
                dp.Field(key="date", label="Date"),
                dp.Field(key="price", label="Price"),
                dp.Field(key="shares", label="Shares"),
                dp.Field(key="total_value", label="Total Value"),
                dp.Field(key="currency_fx_rate", label="Currency Rate"),
                dp.Field(key="total_value_fx_portfolio", label="Total Value Portfolio"),
                dp.Field(key="weighting", label="Weighting"),
                dp.Field(key="portfolio_created", label="Portfolio Created"),
                dp.Field(key="market_share", label="Market Share"),
                dp.Field(key="liquidity", label="Liquidity"),
            ],
            legends=self.LEGENDS,
            formatting=self.FORMATTING,
        )


class AssetPositionPortfolioDisplayConfig(AssetPositionDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    label="Instrument",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="underlying_quote_name", label="Name"),
                        dp.Field(key="underlying_quote_isin", label="ISIN"),
                        dp.Field(key="underlying_quote_ticker", label="Ticker"),
                        dp.Field(key="exchange", label="Exchange", width=Unit.PIXEL(250)),
                    ],
                ),
                dp.Field(
                    label="Valuation",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="weighting", label="Weighting"),
                        dp.Field(key="price", label="Price"),
                        dp.Field(key="shares", label="Shares"),
                        dp.Field(key="currency_fx_rate", label="Currency Rate", show="open"),
                        dp.Field(
                            key="total_value_fx_portfolio",
                            label="Total Value Portfolio",
                            width=Unit.PIXEL(150),
                        ),
                        dp.Field(key="total_value", label="Total Value", show="open"),
                        dp.Field(
                            key="total_value_fx_usd", label="Total Value ($)", width=Unit.PIXEL(150), show="open"
                        ),
                    ],
                ),
                dp.Field(
                    label="Portfolio",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="asset_valuation_date", label="Value Date", width=Unit.PIXEL(100)),
                        dp.Field(
                            key="portfolio_created", label="Portfolio Created", width=Unit.PIXEL(150), show="open"
                        ),
                        dp.Field(key="market_share", label="Market Share", show="open"),
                        dp.Field(key="liquidity", label="Liquidity", show="open"),
                    ],
                ),
            ],
            legends=self.LEGENDS,
            formatting=self.FORMATTING,
        )


class AssetPositionInstrumentDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        base_display = AssetPositionPortfolioDisplayConfig(
            view=self.view, request=self.request, instance=self.instance
        )
        return dp.ListDisplay(
            fields=[
                base_display.get_list_display().fields[0],
                base_display.get_list_display().fields[1],
                dp.Field(
                    label="Portfolio",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="asset_valuation_date", label="Value Date", width=Unit.PIXEL(100)),
                        dp.Field(key="portfolio", label="Portfolio", width=Unit.PIXEL(150)),
                        dp.Field(
                            key="portfolio_created", label="Portfolio Created", width=Unit.PIXEL(150), show="open"
                        ),
                        dp.Field(key="market_share", label="Market Share", show="open"),
                        dp.Field(key="liquidity", label="Liquidity", show="open"),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="is_invested",
                    items=[dp.LegendItem(icon=WBIcon.UNFILTER.icon, label="Not Invested", value=True)],
                )
            ],
        )


class CashPositionPortfolioDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="portfolio_name", label="Portfolio"),
                dp.Field(
                    key="total_value_fx_usd",
                    label="Total Value ($)",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "fontWeight": "bold",
                            },
                        ),
                        dp.FormattingRule(
                            style={"color": WBColor.RED_DARK.value},
                            condition=("<", 0),
                        ),
                    ],
                ),
                dp.Field(key="portfolio_weight", label="Portfolio Weight"),
            ]
        )


class CompositionModelPortfolioPandasDisplayConfig(DisplayViewConfig):
    view: "CompositionModelPortfolioPandasView"

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="underlying_instrument_repr", label="Instrument", width=450),
            dp.Field(
                key=f"group_{self.view.portfolio.id}",
                label=str(self.view.portfolio),
                children=[
                    dp.Field(
                        key=f"shares_{self.view.portfolio.id}",
                        label="Shares",
                        width=150,
                    ),
                    dp.Field(
                        key=f"weighting_{self.view.portfolio.id}",
                        label="Weighting",
                        width=150,
                    ),
                ],
            ),
        ]
        for portfolio in self.view.dependant_portfolios:
            fields.append(
                dp.Field(
                    key=f"group_{portfolio.id}",
                    label=str(portfolio),
                    children=[
                        dp.Field(
                            key=f"shares_{portfolio.id}",
                            label="Shares",
                            width=150,
                        ),
                        dp.Field(
                            key=f"weighting_{portfolio.id}",
                            label="Weighting",
                            width=150,
                        ),
                        dp.Field(
                            key=f"difference_{portfolio.id}",
                            label="Δ",
                            width=75,
                            formatting_rules=[
                                dp.FormattingRule(
                                    condition=("<", 0), style={"fontWeight": "bold", "color": "#FF6961"}
                                ),
                                dp.FormattingRule(condition=(">", 0), style={"fontWeight": "bold", "color": "green"}),
                            ],
                        ),
                    ],
                ),
            )
        return dp.ListDisplay(fields=fields)


class DistributionTableDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="weighting", label="Weighting"),
        ]
        for k, v in reversed(self.view.columns_map.items()):
            fields.append(dp.Field(key=k, label=v))
        return dp.ListDisplay(fields=fields)



# Generated by Django 5.0.14 on 2025-07-30 14:40

from decimal import Decimal
from django.db import migrations, models
from tqdm import tqdm

def migrate_desired_target_weight(apps, schema_editor):
    from wbportfolio.models.orders import Order, OrderProposal
    objs = []
    qs = OrderProposal.objects.filter(orders__isnull=False).distinct()
    for op in tqdm(qs, total=qs.count()):
        for o in op.get_orders():
            o.desired_target_weight = o._target_weight
            objs.append(o)
    Order.objects.bulk_update(objs, ["desired_target_weight"], batch_size=10000)


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0084_orderproposal_min_order_value'),
    ]

    operations = [
        migrations.RunPython(migrate_desired_target_weight)
    ]



from django.apps import AppConfig
from django.db.models.signals import post_migrate
from wbcore.contrib.directory.configurations import configuration_registry


class WbportfolioDirectoryConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "wbportfolio.contrib.company_portfolio"

    def ready(self) -> None:
        from .management import initialize_task

        post_migrate.connect(
            initialize_task,
            dispatch_uid="wbportfolio.contrib.company_portfolio.tasks.initialize_task",
        )

        configuration_registry.DEFAULT_COMPANY_MODEL_VIEWSET = (
            "wbportfolio.contrib.company_portfolio.viewsets.CompanyModelViewSet"
        )
        configuration_registry.DEFAULT_COMPANY_MODEL_SERIALIZER = (
            "wbportfolio.contrib.company_portfolio.serializers.CompanyModelSerializer"
        )
        configuration_registry.DEFAULT_PERSON_MODEL_VIEWSET = (
            "wbportfolio.contrib.company_portfolio.viewsets.PersonModelViewSet"
        )
        configuration_registry.DEFAULT_PERSON_MODEL_SERIALIZER = (
            "wbportfolio.contrib.company_portfolio.serializers.PersonModelSerializer"
        )



from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbfdm.viewsets.configs.buttons.instruments import InstrumentButtonViewConfig

from wbportfolio.models import Product


class ProductButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()

    def get_custom_instance_buttons(self):
        buttons = [
            bt.DropDownButton(
                label="Commission",
                icon=WBIcon.UNFOLD.icon,
                buttons=(
                    bt.WidgetButton(key="claims", label="Claims"),
                    bt.WidgetButton(key="aum", label="AUM per account"),
                ),
            ),
            *InstrumentButtonViewConfig(self.view, self.request, self.instance).get_custom_instance_buttons(),
            bt.DropDownButton(
                label="Risk Management",
                icon=WBIcon.UNFOLD.icon,
                buttons=[
                    bt.WidgetButton(
                        key="risk_rules",
                        label="Rules",
                        icon=WBIcon.CONFIGURE.icon,
                    ),
                    bt.WidgetButton(
                        key="risk_incidents",
                        label="Incidents",
                        icon=WBIcon.WARNING.icon,
                    ),
                ],
            ),
        ]
        if product_id := self.view.kwargs.get("pk", None):
            product = Product.objects.get(id=product_id)
            report_buttons = []
            for report in product.reports.filter(is_active=True):
                tmp_buttons = []
                if primary_version := report.primary_version:
                    tmp_buttons.append(
                        bt.HyperlinkButton(
                            label="Public Report",
                            endpoint=reverse(
                                "public_report:report_version", args=[primary_version.lookup], request=self.request
                            ),
                        ),
                    )
                    if self.request.user.profile.is_internal or self.request.user.is_superuser:
                        tmp_buttons.append(
                            bt.WidgetButton(
                                label="Widget",
                                endpoint=reverse("wbreport:report-detail", args=[report.id], request=self.request),
                            ),
                        )
                if tmp_buttons:
                    report_buttons.append(bt.DropDownButton(label=str(report), buttons=tmp_buttons))
            if report_buttons:
                buttons.append(bt.DropDownButton(label="Reports", buttons=report_buttons))

        return set(buttons)


class ProductCustomerButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(
                key="historical-chart",
                label="Historical Chart",
                icon=WBIcon.STATS.icon,
            )
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from datetime import date
from decimal import Decimal

import numpy as np
import pandas as pd
from django.db.models import (
    DecimalField,
    Expression,
    ExpressionWrapper,
    F,
    FloatField,
    OuterRef,
    Q,
    Subquery,
    Value,
)
from django.db.models.functions import Coalesce
from django.utils.functional import cached_property
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.pandas import fields as pf
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.strings import format_number
from wbfdm.models import Instrument, InstrumentPrice, RelatedInstrumentThroughModel

from wbportfolio.filters import (
    PerformanceComparisonFilter,
    PerformancePandasFilter,
    ProductPerformanceNetNewMoneyFilter,
)
from wbportfolio.models import Product, Trade

from .configs import (
    PerformanceComparisonDisplayConfig,
    PerformanceComparisonEndpointConfig,
    PerformanceComparisonTitleConfig,
    PerformancePandasDisplayConfig,
    PerformancePandasEndpointConfig,
    PerformancePandasTitleConfig,
    ProductPerformanceNetNewMoneyDisplayConfig,
    ProductPerformanceNetNewMoneyEndpointConfig,
    ProductPerformanceNetNewMoneyTitleConfig,
)


class PerformancePandasView(InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbportfolio:performances"

    queryset = Product.objects.all()

    filterset_class = PerformancePandasFilter

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="computed_str", label="Title"),
            pf.CharField(key="currency_symbol", label="Currency Key"),
            pf.BooleanField(key="is_invested", label="Invested"),
            pf.FloatField(key="sum_shares2", label="Shares"),
            pf.FloatField(
                key="n1",
                label="N1",
                precision=2,
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="n2",
                label="N2",
                precision=2,
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="diff",
                label="Difference",
                precision=2,
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(key="perf", label="Performance", precision=2, percent=True),
            pf.FloatField(
                key="n1_usd",
                label="N1 (usd)",
                precision=2,
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(
                key="n2_usd",
                label="N2 (usd)",
                precision=2,
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(
                key="diff_usd",
                label="Difference (usd)",
                precision=2,
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(key="perf_usd", label="Performance (usd)", precision=2, percent=True),
        )
    )
    display_config_class = PerformancePandasDisplayConfig
    title_config_class = PerformancePandasTitleConfig
    endpoint_config_class = PerformancePandasEndpointConfig

    search_fields = ["computed_str"]
    ordering_fields = ["computed_str", "n1", "n2", "diff", "perf", "n1_usd", "n2_usd", "diff_usd", "perf_usd"]

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        return {
            "n1_usd": {"Σ": format_number(df["n1_usd"].sum())},
            "n2_usd": {"Σ": format_number(df["n2_usd"].sum())},
            "diff_usd": {"Σ": format_number(df["diff_usd"].sum())},
        }

    def get_queryset(self):
        d1, d2 = get_date_interval_from_request(self.request, exclude_weekend=True)
        if d1 and d2:
            return (
                super()
                .get_queryset()
                .annotate(
                    d1=Subquery(
                        InstrumentPrice.objects.filter(
                            calculated=False, instrument=OuterRef("pk"), date__gte=d1, date__lte=d2
                        )
                        .order_by("date")
                        .values("date")[:1]
                    ),
                    d2=Subquery(
                        InstrumentPrice.objects.filter(
                            calculated=False, instrument=OuterRef("pk"), date__gte=d1, date__lte=d2
                        )
                        .order_by("-date")
                        .values("date")[:1]
                    ),
                    fx_rate_n1=Coalesce(
                        Subquery(
                            CurrencyFXRates.objects.filter(currency=OuterRef("currency"), date=OuterRef("d1")).values(
                                "value"
                            )[:1]
                        ),
                        Decimal(1),
                    ),
                    fx_rate_n2=Coalesce(
                        Subquery(
                            CurrencyFXRates.objects.filter(currency=OuterRef("currency"), date=OuterRef("d2")).values(
                                "value"
                            )[:1]
                        ),
                        Decimal(1),
                    ),
                    outstanding_shares_n1=Coalesce(
                        InstrumentPrice.subquery_closest_value(
                            "outstanding_shares",
                            instrument_pk_name="pk",
                            date_name="d1",
                            order_by="calculated",
                            calculated_filter_value=None,
                            date_lookup="exact",
                        ),
                        Decimal(0),
                    ),
                    outstanding_shares_n2=Coalesce(
                        InstrumentPrice.subquery_closest_value(
                            "outstanding_shares",
                            instrument_pk_name="pk",
                            date_name="d2",
                            order_by="calculated",
                            calculated_filter_value=None,
                            date_lookup="exact",
                        ),
                        Decimal(0),
                    ),
                    net_value_n1=InstrumentPrice.subquery_closest_value(
                        "net_value", instrument_pk_name="pk", date_name="d1", date_lookup="exact"
                    ),
                    net_value_n2=InstrumentPrice.subquery_closest_value(
                        "net_value", instrument_pk_name="pk", date_name="d2", date_lookup="exact"
                    ),
                )
                .filter(d1__isnull=False, d2__isnull=False)
            )
        return Product.objects.none()

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame(
            queryset.values(
                "id",
                "computed_str",
                "currency__symbol",
                "d1",
                "d2",
                "fx_rate_n1",
                "fx_rate_n2",
                "outstanding_shares_n1",
                "outstanding_shares_n2",
                "net_value_n1",
                "net_value_n2",
            )
        ).rename(columns={"currency__symbol": "currency_symbol"})
        performance_by = self.request.GET.get("performance_by", PerformancePandasFilter.PerformanceBy.PRICE.name)
        if not df.empty:
            if performance_by == PerformancePandasFilter.PerformanceBy.NNM.name:
                df["n1"] = df["outstanding_shares_n1"] * df["net_value_n1"]
                df["n2"] = df["outstanding_shares_n2"] * df["net_value_n1"]
                df["n1_usd"] = df["n1"] / df["fx_rate_n1"]
                df["n2_usd"] = df["n2"] / df["fx_rate_n2"]
            elif performance_by == PerformancePandasFilter.PerformanceBy.AUM.name:
                df["n1"] = df["outstanding_shares_n1"] * df["net_value_n1"]
                df["n2"] = df["outstanding_shares_n2"] * df["net_value_n2"]
                df["n1_usd"] = df["n1"] / df["fx_rate_n1"]
                df["n2_usd"] = df["n2"] / df["fx_rate_n2"]

            elif performance_by == PerformancePandasFilter.PerformanceBy.PRICE.name:
                df["n1"] = df["net_value_n1"]
                df["n2"] = df["net_value_n2"]
                df["n1_usd"] = df["n1"] / df["fx_rate_n1"]
                df["n2_usd"] = df["n2"] / df["fx_rate_n2"]
            else:
                raise ValueError("You need to specify a performance by filter")

            df = df.ffill()
            number_columns = df.columns.difference(["id", "computed_str", "currency_symbol", "d1", "d2"])
            df[number_columns] = df[number_columns].astype(float)
            df["diff"] = df["n2"] - df["n1"]
            df["diff_usd"] = df["n2_usd"] - df["n1_usd"]
            df["perf"] = df["diff"].div(df["n1"])
            df["perf_usd"] = df["diff_usd"].div(df["n1_usd"])
        return df.replace(np.inf, 0)


class ProductPerformanceNetNewMoneyListViewSet(InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbportfolio:product-performance-net-new-money"

    filterset_class = ProductPerformanceNetNewMoneyFilter
    queryset = Product.objects.all()

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="computed_str", label="Title"),
            pf.CharField(key="currency_symbol", label="Currency Key"),
            pf.BooleanField(key="is_invested", label="Invested"),
            pf.FloatField(
                key="net_money",
                label="Net New Money",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="net_negative_money",
                label="Net Negative Money",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="net_positive_money",
                label="Net Positive Money",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="net_money_usd",
                label="Net Money ($)",
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(
                key="net_negative_money_usd",
                label="Net Negative Money ($)",
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(
                key="net_positive_money_usd",
                label="Net Positive Money ($)",
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
        )
    )

    search_fields = ("computed_str", "isin", "ticker")
    ordering_fields = (
        "computed_str",
        "net_negative_money",
        "net_negative_money_usd",
        "net_positive_money",
        "net_positive_money_usd",
        "net_money",
        "net_money_usd",
    )
    ordering = ["computed_str"]

    display_config_class = ProductPerformanceNetNewMoneyDisplayConfig
    title_config_class = ProductPerformanceNetNewMoneyTitleConfig
    endpoint_config_class = ProductPerformanceNetNewMoneyEndpointConfig

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        return {
            "net_money_usd": {"Σ": format_number(df.net_money_usd.sum())},
            "net_negative_money_usd": {"Σ": format_number(df.net_negative_money_usd.sum())},
            "net_positive_money_usd": {"Σ": format_number(df.net_positive_money_usd.sum())},
        }

    @cached_property
    def latest_rate_date(self) -> date:
        try:
            return CurrencyFXRates.objects.latest("date").date
        except CurrencyFXRates.DoesNotExist:
            return date.today()

    def get_queryset(self):
        d1, d2 = get_date_interval_from_request(self.request, exclude_weekend=True)
        d1 = (
            Trade.objects.earliest("transaction_date").transaction_date
            if not d1 and Trade.objects.all().exists()
            else d1
        )
        d2 = (
            Trade.objects.latest("transaction_date").transaction_date
            if not d2 and Trade.objects.all().exists()
            else d2
        )
        products = Product.objects.all().annotate(currency_symbol=F("currency__symbol"))
        if d1 and d2:
            products = products.annotate(
                sum_shares=Trade.subquery_shares_per_underlying_instrument(d1),
                net_money=Trade.subquery_net_money(d1, d2),
                net_negative_money=Trade.subquery_net_money(d1, d2, only_negative=True),
                net_positive_money=Trade.subquery_net_money(d1, d2, only_positive=True),
                fx2=CurrencyFXRates.get_fx_rates_subquery(
                    self.latest_rate_date, currency="currency", lookup_expr="exact"
                ),
                net_money_usd=ExpressionWrapper(F("fx2") * F("net_money"), output_field=FloatField()),
                net_negative_money_usd=ExpressionWrapper(
                    F("fx2") * F("net_negative_money"), output_field=FloatField()
                ),
                net_positive_money_usd=ExpressionWrapper(
                    F("fx2") * F("net_positive_money"), output_field=FloatField()
                ),
            )
        else:
            products = products.annotate(
                sum_shares=Value(0.0),
                net_money=Value(0.0),
                net_negative_money=Value(0.0),
                net_positive_money=Value(0.0),
                fx2=Value(0.0),
                net_money_usd=Value(0.0),
                net_negative_money_usd=Value(0.0),
                net_positive_money_usd=Value(0.0),
            )
        return products.select_related("currency")


class PerformanceComparisonPandasView(InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbportfolio:performancescomparison"

    queryset = Product.objects.all()
    filterset_class = PerformanceComparisonFilter

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="computed_str", label="Title"),
            pf.CharField(key="benchmark_computed_str", label="Compared benchmark"),
            pf.CharField(key="currency_symbol", label="Currency Key"),
            pf.FloatField(
                key="instrument_last_valuation_price",
                label="N1",
                precision=2,
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.DateField(key="inception_date", label="Inception Date"),
            pf.DateField(key="last_valuation_date", label="Last Date Date"),
            pf.FloatField(key="perf_last_day", label="daily", precision=2, percent=True),
            pf.FloatField(key="perf_month_to_date", label="monthly", precision=2, percent=True),
            pf.FloatField(key="perf_year_to_date", label="yearly", precision=2, percent=True),
            pf.FloatField(key="perf_inception", label="inception", precision=2, percent=True),
            pf.FloatField(key="perf_between_dates", label="Perf", precision=2, percent=True),
        )
    )
    display_config_class = PerformanceComparisonDisplayConfig
    title_config_class = PerformanceComparisonTitleConfig
    endpoint_config_class = PerformanceComparisonEndpointConfig

    search_fields = ["computed_str"]
    ordering = ["computed_str"]
    ordering_fields = [
        "computed_str",
        "instrument_last_valuation_price",
        "last_valuation_date",
        "perf_last_day",
        "perf_month_to_date",
        "perf_year_to_date",
        "perf_inception",
        "last_valuation_date",
        "perf_between_dates",
    ]

    @property
    def dates(self):
        if (
            (dates := get_date_interval_from_request(self.request, exclude_weekend=True, date_range_fieldname="dates"))
            and (d1 := dates[0])
            and (d2 := dates[1])
        ):
            return d1, d2

    @property
    def benchmark_label(self) -> ExpressionWrapper:
        comparison_instrument_id = self.request.GET.get("comparison_instrument", None)
        compare_primary_benchmark = self.request.GET.get("compare_primary_benchmark", "false") == "true"
        if comparison_instrument_id or compare_primary_benchmark:
            return (
                Value(comparison_instrument_id)
                if comparison_instrument_id
                else Subquery(
                    RelatedInstrumentThroughModel.objects.filter(
                        instrument=OuterRef("pk"),
                        is_primary=True,
                        related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.BENCHMARK,
                    ).values("related_instrument")[:1]
                )
            )

    def get_queryset(self):
        def _prices_annotation_expressions(
            prefix: str = "instrument", instrument_label: str = "pk", initial_price_label: str = "share_price"
        ) -> dict[str, Expression]:
            """
            Utility function to prepare annotatione expression for the instrument or benchmark
            Args:
                prefix: prefix to use for the annotated field names
                instrument_label: OuterRef instrument label
                initial_price_label: OuterRef initial instrument price
            Returns:
                A dictionary of annotable expressions
            """
            res = {
                f"{prefix}_last_valuation_price": Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef(instrument_label),
                            calculated=False,
                            date=OuterRef("last_valuation_date"),
                        )
                        .exclude(net_value=0)
                        .values("net_value")[:1]
                    ),
                    ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                ),
                f"_{prefix}_initial_price": Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef(instrument_label), calculated=False, date=OuterRef("inception_date")
                        )
                        .exclude(net_value=0)
                        .values("net_value")[:1]
                    ),
                    ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                ),
                f"_{prefix}_valuation_price_t1": Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef(instrument_label), calculated=False, date=OuterRef("_t1")
                        )
                        .exclude(net_value=0)
                        .values("net_value")[:1]
                    ),
                    ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                ),
                f"_{prefix}_valuation_price_m1": Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef(instrument_label), calculated=False, date=OuterRef("_m1")
                        )
                        .exclude(net_value=0)
                        .values("net_value")[:1]
                    ),
                    ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                ),
                f"_{prefix}_valuation_price_y1": Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef(instrument_label),
                            calculated=False,
                            date=OuterRef("_y1"),
                        )
                        .exclude(net_value=0)
                        .values("net_value")[:1]
                    ),
                    ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                ),
                f"_{prefix}_perf_last_day": F(f"{prefix}_last_valuation_price") / F(f"_{prefix}_valuation_price_t1")
                - 1,
                f"_{prefix}_perf_month_to_date": F(f"{prefix}_last_valuation_price")
                / F(f"_{prefix}_valuation_price_m1")
                - 1,
                f"_{prefix}_perf_year_to_date": F(f"{prefix}_last_valuation_price")
                / F(f"_{prefix}_valuation_price_y1")
                - 1,
                f"_{prefix}_perf_inception": F(f"{prefix}_last_valuation_price") / F(f"_{prefix}_initial_price") - 1,
            }
            if self.dates:
                # If dates is available, we also annotate the performance between the specified dates range
                res.update(
                    {
                        f"_{prefix}_last_valuation_price_d1": Coalesce(
                            Subquery(
                                InstrumentPrice.objects.filter(
                                    instrument=OuterRef(instrument_label), calculated=False, date__lte=self.dates[0]
                                )
                                .order_by("-date")
                                .values("net_value")[:1]
                            ),
                            ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                        ),
                        f"_{prefix}_last_valuation_price_d2": Coalesce(
                            Subquery(
                                InstrumentPrice.objects.filter(
                                    instrument=OuterRef(instrument_label),
                                    calculated=False,
                                    date__lte=self.dates[1],
                                )
                                .order_by("-date")
                                .values("net_value")[:1]
                            ),
                            ExpressionWrapper(F(initial_price_label), output_field=DecimalField()),
                        ),
                        f"_{prefix}_perf_between_dates": F(f"_{prefix}_last_valuation_price_d2")
                        / F(f"_{prefix}_last_valuation_price_d1")
                        - 1,
                    }
                )
            else:
                res[f"_{prefix}_perf_between_dates"] = Value(
                    Decimal(0)
                )  # Ensure the key is available to avoid keyerror during evaluation
            return res

        qs = (
            super()
            .get_queryset()
            .annotate(
                currency_symbol=F("currency__symbol"),
                _t1=Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef("pk"), calculated=False, date__lt=OuterRef("last_valuation_date")
                        )
                        .order_by("-date")
                        .values("date")[:1]
                    ),
                    F("inception_date"),
                ),  # t-1 date
                _m1=Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            Q(instrument=OuterRef("pk"))
                            & Q(calculated=False)
                            & ~(
                                Q(date__gte=OuterRef("last_valuation_date"))
                                | (
                                    Q(date__month=OuterRef("last_valuation_date__month"))
                                    & Q(date__year=OuterRef("last_valuation_date__year"))
                                )
                            )
                        )
                        .order_by("-date")
                        .values("date")[:1]
                    ),
                    F("inception_date"),
                ),  # latest previous month date
                _y1=Coalesce(
                    Subquery(
                        InstrumentPrice.objects.filter(
                            instrument=OuterRef("pk"),
                            calculated=False,
                            date__year=OuterRef("last_valuation_date__year") - 1,
                        )
                        .order_by("-date")
                        .values("date")[:1]
                    ),
                    F("inception_date"),
                ),  # latest previous year date
                **_prices_annotation_expressions(),
            )
        )
        if benchmark_label := self.benchmark_label:
            # If benchmark is provided, we annotate their respective price depending on the outer reference POI
            qs = qs.annotate(
                _benchmark_id=benchmark_label,
                benchmark_computed_str=Subquery(
                    Instrument.objects.filter(id=OuterRef("_benchmark_id")).values("computed_str")[:1]
                ),
                benchmark_issue_price=Subquery(
                    Instrument.objects.filter(id=OuterRef("_benchmark_id")).values("issue_price")[:1]
                ),
                _benchmark_last_valuation_price=Subquery(
                    InstrumentPrice.objects.filter(
                        instrument=OuterRef("_benchmark_id"), calculated=False, date=OuterRef("last_valuation_date")
                    ).values("net_value")[:1]
                ),
                **_prices_annotation_expressions(
                    prefix="benchmark", instrument_label="_benchmark_id", initial_price_label="benchmark_issue_price"
                ),
            )
        else:
            qs = qs.annotate(
                benchmark_computed_str=Value(""),
                **{
                    k: Value(Decimal(0.0))
                    for k in [
                        "_benchmark_perf_between_dates",
                        "_benchmark_perf_last_day",
                        "_benchmark_perf_month_to_date",
                        "_benchmark_perf_year_to_date",
                        "_benchmark_perf_inception",
                    ]
                },
            )  # Ensure the key is available to avoid keyerror during evaluation
        return qs.annotate(
            perf_last_day=F("_instrument_perf_last_day") - Coalesce(F("_benchmark_perf_last_day"), Value(Decimal(0))),
            perf_month_to_date=F("_instrument_perf_month_to_date")
            - Coalesce(F("_benchmark_perf_month_to_date"), Value(Decimal(0))),
            perf_year_to_date=F("_instrument_perf_year_to_date")
            - Coalesce(F("_benchmark_perf_year_to_date"), Value(Decimal(0))),
            perf_inception=F("_instrument_perf_inception")
            - Coalesce(F("_benchmark_perf_inception"), Value(Decimal(0))),
            perf_between_dates=F("_instrument_perf_between_dates")
            - Coalesce(F("_benchmark_perf_between_dates"), Value(Decimal(0))),
        )



from datetime import date

from django.db import models

from wbportfolio.models.products import Product


class DataBackendMixin:
    def is_object_valid(self, obj: models.Model) -> bool:
        return super().is_object_valid(obj) and obj.is_active_at_date(date.today()) and obj.isin

    def get_default_queryset(self):
        return Product.objects.filter(bank=self.ubs_bank, isin__isnull=False)

    def get_provider_id(self, obj: models.Model) -> str:
        return obj.isin



import random

import factory
from faker import Faker

from wbportfolio.models import Fees

faker = Faker()


class FeesFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Fees

    currency_fx_rate = 1.0
    fee_date = factory.Faker("date_object")
    total_value = factory.LazyAttribute(lambda o: random.randint(1, 1000))
    transaction_subtype = factory.Faker("random_element", elements=[x[0] for x in Fees.Type.choices])
    product = factory.SubFactory("wbportfolio.factories.products.ProductFactory")
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")



from collections import defaultdict
from contextlib import suppress
from datetime import datetime
from decimal import Decimal
from itertools import chain
from typing import Any, Dict, List, Optional

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from wbcore.contrib.authentication.authentication import User
from wbcore.contrib.currency.import_export.handlers import CurrencyImportHandler
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.io.exceptions import DeserializationError, SkipImportError
from wbcore.contrib.io.imports import ImportExportHandler
from wbcore.contrib.notifications.dispatch import send_notification
from wbfdm.import_export.handlers.instrument import InstrumentImportHandler
from wbfdm.import_export.handlers.instrument_price import InstrumentPriceImportHandler
from wbfdm.models.exchanges import Exchange

from wbportfolio.models.roles import PortfolioRole


class AssetPositionImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbportfolio.AssetPosition"
    PRICE_DATE_TIMEDELTA: int = 7
    MAX_PRICE_DATE_TIMEDELTA: int = 360

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)
        self.instrument_price_handler = InstrumentPriceImportHandler(self.import_source)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _deserialize(self, data: Dict[str, Any]):
        from wbportfolio.models import Portfolio

        portfolio_data = data.pop("portfolio", None)
        underlying_quote_data = data.pop("underlying_quote", data.pop("underlying_instrument", None))
        if "currency" in data:
            currency = self.currency_handler.process_object(data["currency"], read_only=True, raise_exception=False)[0]
            # we do not support GBX in our instrument table
            if currency.key == "GBX":
                currency = Currency.objects.get(key="GBP")
                if "initial_price" in data:
                    data["initial_price"] /= 1000
            data["currency"] = currency
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()
        if data.get("asset_valuation_date", None):
            data["asset_valuation_date"] = datetime.strptime(data["asset_valuation_date"], "%Y-%m-%d").date()
        else:
            data["asset_valuation_date"] = data["date"]

        if exchange_data := data.pop("exchange", None):
            sanitized_dict = {k: v for k, v in exchange_data.items() if v is not None}
            if sanitized_dict:
                data["exchange"] = Exchange.dict_to_model(sanitized_dict)

        data["portfolio"] = Portfolio._get_or_create_portfolio(self.instrument_handler, portfolio_data)
        data["underlying_quote"] = self.instrument_handler.process_object(
            underlying_quote_data, only_security=False, read_only=True
        )[0]

        # Ensure that for shares and weighting, a None value default to 0
        if "initial_shares" in data and data["initial_shares"] is None:
            data["initial_shares"] = Decimal(0)
        if "weighting" in data and data["weighting"] is None:
            data["weighting"] = Decimal(0)

        # if the initial price is not provided, we try to get it directly from the dataloader
        if data.get("initial_price") is None:
            try:
                data["initial_price"] = data["underlying_quote"].get_price(
                    data["date"], price_date_timedelta=self.PRICE_DATE_TIMEDELTA
                )
            except ValueError:
                # If we cannot find a price with the default timedelta, we try with a bigger range (in case the instrument stopped trading for some time for instance)
                try:
                    data["initial_price"] = data["underlying_quote"].get_price(
                        data["date"], price_date_timedelta=self.MAX_PRICE_DATE_TIMEDELTA
                    )
                except ValueError as e:
                    raise DeserializationError("Price not provided but can not be found automatically") from e

        # number type deserialization and sanitization
        # ensure the provided Decimal field are of type Decimal
        decimal_fields = ["initial_currency_fx_rate", "initial_price", "initial_shares", "weighting"]
        for field in decimal_fields:
            if (value := data.get(field, None)) is not None:
                data[field] = Decimal(value)

        if data["weighting"] == 0:
            raise SkipImportError("exclude position whose weight is 0")

    def _process_raw_data(self, data: Dict[str, Any]):
        if prices := data.get("prices", None):
            self.import_source.log += "Instrument Prices found: Importing"
            self.instrument_price_handler.process({"data": prices})

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        self.import_source.log += "\nTrying to get asset position instance"
        position = data["portfolio"].assets.filter(date=data["date"], underlying_quote=data["underlying_quote"])
        if position.exists():
            self.import_source.log += "\nAsset Position found."
            if position.count() > 1:
                position = position.filter(is_estimated=False)
            if position.count() > 1:
                raise ValueError(f'We should find only one Assetposition:{position.values_list("id", flat=True)}')
            return position.first()
        self.import_source.log += "\nAsset Position not found. A new one will be created"
        return None

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        instance = self.model(
            portfolio=data["portfolio"],
            underlying_quote=data["underlying_quote"],
            date=data["date"],
            asset_valuation_date=data["asset_valuation_date"],
            weighting=data.get("weighting", None),
            initial_price=data["initial_price"],
            initial_shares=data.get("initial_shares", None),
            initial_currency_fx_rate=data.get("initial_currency_fx_rate", 1),
            import_source=self.import_source,
            currency=data.get("currency", None),
        )

        return self._save_object(instance, **kwargs)

    def _save_object(self, _object, **kwargs):
        _object.underlying_quote_price = (
            None  # detech possibly already attached instrument price to retrigger the save mechanism
        )
        _object.save(create_underlying_quote_price_if_missing=True)
        return _object

    def _post_processing_objects(
        self,
        created_objs: List[models.Model],
        modified_objs: List[models.Model],
        unmodified_objs: List[models.Model],
    ):
        from wbportfolio.models.portfolio import trigger_portfolio_change_as_task

        portfolio_to_resynch = defaultdict(set)
        imported_ids = defaultdict(set)
        for obj in chain(created_objs, modified_objs, unmodified_objs):
            portfolio_to_resynch[obj.portfolio].add(obj.date)
            imported_ids[obj.portfolio].add(obj.id)

        for portfolio, dates in portfolio_to_resynch.items():
            # We remove leftovers positions from wrongly imported file
            leftovers_positions = self.model.objects.filter(portfolio=portfolio, date__in=dates)
            for obj_id in imported_ids[portfolio]:
                leftovers_positions = leftovers_positions.exclude(id=obj_id)
            for position in leftovers_positions:
                position.delete()
            for val_date in sorted(dates):
                trigger_portfolio_change_as_task.delay(portfolio.id, val_date, fix_quantization=True)

            # check if portfolio as custodian
            latest_date = max(dates)
            with suppress(ObjectDoesNotExist):
                if (custodian_rel := portfolio.dependency_through.get(type="CUSTODIAN")) and portfolio.assets.latest(
                    "date"
                ).date >= latest_date:
                    custodian_portfolio = custodian_rel.dependency_portfolio
                    differences = portfolio.check_related_portfolio_at_date(latest_date, custodian_portfolio)
                    if differences.exists():
                        for user in User.objects.filter(
                            profile_id__in=PortfolioRole.portfolio_managers().values_list("person", flat=True)
                        ):
                            send_notification(
                                code="wbportfolio.portfolio.check_custodian_portfolio",
                                title="There is a discrepency between two portfolios",
                                body=f"There has been a discrepency between two portfolios: {portfolio} and {custodian_portfolio}.",
                                user=user,
                                reverse_name="wbportfolio:portfolio-modelcompositionpandas-list",
                                reverse_args=[portfolio.id],
                            )



from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.search import TrigramSimilarity
from django.db import models
from wbcore.contrib.directory.models import Company
from wbcore.models import WBModel


class Custodian(WBModel):
    name = models.CharField(
        max_length=255,
        unique=True,
        verbose_name="Name",
        help_text="The Name of the Custodian.",
    )
    company = models.ForeignKey(
        "directory.Company", null=True, blank=True, on_delete=models.SET_NULL, related_name="custodians"
    )
    mapping = ArrayField(
        base_field=models.CharField(max_length=255),
        default=list,
        verbose_name="Custodian Mapping",
        help_text="Each custodian of this mapping will be assigned to this custodian.",
    )

    @staticmethod
    def get_unspecified_bank_name():
        return "<NO BANK SPECIFIED>"

    @classmethod
    def get_unspecified_bank(cls):
        return cls.get_by_mapping(cls.get_unspecified_bank_name())

    @classmethod
    def get_by_mapping(cls, mapping: str, use_similarity=False, create_missing=True):
        similarity_score = 0.7
        lower_mapping = mapping.lower()
        try:
            return cls.objects.get(mapping__contains=[lower_mapping])
        except cls.DoesNotExist:
            if use_similarity:
                similar_custodians = cls.objects.annotate(
                    similarity_score=TrigramSimilarity("name", lower_mapping)
                ).filter(similarity_score__gt=similarity_score)
                if similar_custodians.count() == 1:
                    custodian = similar_custodians.first()
                    print(f"find similar custodian {lower_mapping} -> {custodian.name}")  # noqa: T201
                    custodian.mapping.append(lower_mapping)
                    custodian.save()
                    return custodian
                else:
                    similar_companies = Company.objects.annotate(
                        similarity_score=TrigramSimilarity("name", lower_mapping)
                    ).filter(similarity_score__gt=similarity_score)
                    if similar_companies.count() == 1:
                        print(  # noqa: T201
                            f"Find similar company {lower_mapping} -> {similar_companies.first().name}"
                        )  # noqa: T201
                        return cls.objects.create(
                            name=lower_mapping, mapping=[lower_mapping], company=similar_companies.first()
                        )
        if create_missing:
            return cls.objects.create(name=lower_mapping, mapping=[lower_mapping])

    def merge(self, custodian):
        if self != custodian:
            custodian.trades.update(custodian=self)
            self.mapping = list(set(self.mapping + custodian.mapping))
            if not self.company and custodian.company:
                self.company = custodian.company
            self.save()
            custodian.delete()

    def split_off(self, mapping):
        if mapping in self.mapping:
            self.mapping.remove(mapping)
            self.save()
            custodian = Custodian.objects.create(name=mapping, mapping=[mapping])
            self.trades.filter(bank=mapping).update(custodian=custodian)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "Custodian"
        verbose_name_plural = "Custodians"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:custodian"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:custodianrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"



from datetime import timedelta
from decimal import Decimal

from celery import shared_task
from django.db import models
from django.db.models import (
    Case,
    DateField,
    ExpressionWrapper,
    F,
    OuterRef,
    Q,
    Subquery,
    Sum,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import post_save
from django.dispatch import receiver
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.signals import pre_merge
from wbcore.signals.models import pre_collection
from wbfdm.models import Instrument
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.signals import add_instrument_to_investable_universe

from wbportfolio.import_export.handlers.trade import TradeImportHandler
from wbportfolio.models.custodians import Custodian

from .transactions import TransactionMixin


class ValidCustomerTradeManager(models.Manager):
    def __init__(self, without_internal_trade: bool = False):
        self.without_internal_trade = without_internal_trade
        super().__init__()

    def get_queryset(self):
        qs = (
            super()
            .get_queryset()
            .filter(
                transaction_subtype__in=[Trade.Type.SUBSCRIPTION, Trade.Type.REDEMPTION],
                marked_for_deletion=False,
                pending=False,
            )
        )
        if self.without_internal_trade:
            qs = qs.exclude(marked_as_internal=True)
        return qs


class Trade(TransactionMixin, ImportMixin, models.Model):
    import_export_handler_class = TradeImportHandler

    TRADE_WINDOW_INTERVAL = 7

    class Type(models.TextChoices):
        REBALANCE = "REBALANCE", "Rebalance"
        DECREASE = "DECREASE", "Decrease"
        INCREASE = "INCREASE", "Increase"
        SUBSCRIPTION = "SUBSCRIPTION", "Subscription"
        REDEMPTION = "REDEMPTION", "Redemption"
        BUY = "BUY", "Buy"
        SELL = "SELL", "Sell"
        NO_CHANGE = "NO_CHANGE", "No Change"  # default transaction subtype if weighing is 0

    transaction_subtype = models.CharField(
        max_length=32, default=Type.BUY, choices=Type.choices, verbose_name="Trade Type"
    )
    transaction_date = models.DateField(
        verbose_name="Trade Date",
        help_text="The date that this transaction was traded.",
    )
    book_date = models.DateField(
        verbose_name="Trade Date",
        help_text="The date that this transaction was booked.",
    )
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal("0.0"),
        help_text="The number of shares that were traded.",
        verbose_name="Shares",
    )

    weighting = models.DecimalField(
        max_digits=9,
        decimal_places=8,
        default=Decimal(0),
        help_text="The weight to be multiplied against the target",
        verbose_name="Weight",
    )
    claimed_shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal(0),
        help_text="The number of shares that were claimed.",
        verbose_name="Claimed Shares",
    )
    diff_shares = models.GeneratedField(
        expression=F("shares") - F("claimed_shares"),
        output_field=models.DecimalField(max_digits=15, decimal_places=4),
        db_persist=True,
    )
    internal_trade = models.OneToOneField(
        "wbportfolio.Trade",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="internal_subscription_redemption_trade",
    )
    marked_for_deletion = models.BooleanField(
        default=False,
        help_text="If this is checked, then the trade is supposed to be deleted.",
        verbose_name="To be deleted",
    )
    marked_as_internal = models.BooleanField(
        default=False,
        help_text="If this is checked, then this subscription or redemption is considered internal and will not be considered in any AUM computation",
        verbose_name="Internal",
    )
    pending = models.BooleanField(default=False)
    exclude_from_history = models.BooleanField(default=False)
    bank = models.CharField(
        max_length=255,
        help_text="The bank/counterparty/custodian the trade went through.",
        verbose_name="Counterparty",
    )
    custodian = models.ForeignKey(
        "wbportfolio.Custodian", null=True, blank=True, on_delete=models.SET_NULL, related_name="trades"
    )
    register = models.ForeignKey(
        to="wbportfolio.Register",
        null=True,
        blank=True,
        related_name="trades",
        on_delete=models.PROTECT,
    )
    external_id = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="An external identifier that was supplied.",
        verbose_name="External Identifier",
    )
    external_id_alternative = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="A second external identifier that was supplied.",
        verbose_name="Alternative External Identifier",
    )

    # Manager
    objects = models.Manager()
    valid_customer_trade_objects = ValidCustomerTradeManager()
    valid_external_customer_trade_objects = ValidCustomerTradeManager(without_internal_trade=True)

    @property
    def product(self):
        from wbportfolio.models.products import Product

        try:
            return Product.objects.get(id=self.underlying_instrument.id)
        except Product.DoesNotExist:
            return None

    class Meta:
        verbose_name = "Trade"
        verbose_name_plural = "Trades"
        indexes = [
            models.Index(fields=["underlying_instrument", "transaction_date"]),
            models.Index(fields=["portfolio", "underlying_instrument", "transaction_date"]),
        ]
        constraints = [
            models.CheckConstraint(
                check=models.Q(marked_as_internal=False)
                | (
                    models.Q(marked_as_internal=True)
                    & models.Q(transaction_subtype__in=["REDEMPTION", "SUBSCRIPTION"])
                ),
                name="marked_as_internal_only_for_subred",
            ),
            models.CheckConstraint(
                check=models.Q(internal_trade__isnull=True)
                | (
                    models.Q(internal_trade__isnull=False)
                    & models.Q(transaction_subtype__in=["REDEMPTION", "SUBSCRIPTION"])
                ),
                name="internal_trade_set_only_for_subred",
            ),
        ]
        # notification_email_template = "portfolio/email/trade_notification.html"

    def save(self, *args, **kwargs):
        self.pre_save()
        if abs(self.weighting) < 10e-6:
            self.weighting = Decimal("0")
        if not self.price:
            # we try to get the price if not provided directly from the underlying instrument
            self.price = self.get_price()

        if not self.custodian and self.bank:
            self.custodian = Custodian.get_by_mapping(self.bank)

        if self.transaction_subtype is None:
            # if subtype not provided, we extract it automatically from the existing data.
            self._set_type()
        if self.id and hasattr(self, "claims"):
            self.claimed_shares = self.claims.filter(status="APPROVED").aggregate(s=Sum("shares"))["s"] or Decimal(0)
        if self.internal_trade:
            self.marked_as_internal = True
        if not self.value_date:
            self.value_date = self.transaction_date
        if not self.book_date:
            self.book_date = self.transaction_date
        super().save(*args, **kwargs)

    def _set_type(self):
        if self.weighting == 0:
            self.transaction_subtype = Trade.Type.NO_CHANGE
        if self.underlying_instrument.instrument_type.key == "product":
            if self.shares is not None:
                if self.shares > 0:
                    self.transaction_subtype = Trade.Type.SUBSCRIPTION
                elif self.shares < 0:
                    self.transaction_subtype = Trade.Type.REDEMPTION
        elif self.weighting is not None:
            if self.weighting > 0:
                self.transaction_subtype = Trade.Type.INCREASE
            elif self.weighting < 0:
                self.transaction_subtype = Trade.Type.DECREASE
        else:
            self.transaction_subtype = Trade.Type.REBALANCE

    def get_price(self) -> Decimal:
        try:
            return self.underlying_instrument.get_price(self.transaction_date)
        except ValueError:
            return Decimal("0")

    def delete(self, **kwargs):
        pre_collection.send(sender=self.__class__, instance=self)
        super().delete(**kwargs)

    def __str__(self):
        ticker = f"{self.underlying_instrument.ticker}:" if self.underlying_instrument.ticker else ""
        return f"{ticker}{self.shares} ({self.bank})"

    def get_alternative_valid_trades(self, share_delta: float = 0):
        return Trade.objects.filter(
            Q(underlying_instrument=self.underlying_instrument)
            & Q(portfolio=self.portfolio)
            & (
                Q(transaction_date__gte=self.transaction_date - timedelta(days=self.TRADE_WINDOW_INTERVAL))
                & Q(transaction_date__lte=self.transaction_date + timedelta(days=self.TRADE_WINDOW_INTERVAL))
            )
            & Q(transaction_subtype=self.transaction_subtype)
            & Q(shares__gte=self.shares * Decimal(1 - share_delta))
            & Q(shares__lte=self.shares * Decimal(1 + share_delta))
            & Q(marked_for_deletion=False)
            & Q(claims__isnull=True)
            & Q(pending=False)
        ).exclude(id=self.id)

    @property
    def is_claimable(self) -> bool:
        return self.is_customer_trade and not self.marked_for_deletion and not self.pending

    @property
    def is_customer_trade(self) -> bool:
        return self.transaction_subtype in [Trade.Type.REDEMPTION.name, Trade.Type.SUBSCRIPTION.name]

    @classmethod
    def subquery_shares_per_underlying_instrument(
        cls, val_date, underlying_instrument_name="pk", only_customer_trade=True
    ):
        """Returns a Subquery that returns the shares at a certain price date
        or 0

        Arguments:
            val_date {datetime.date} -- The  date that is used to determine which tradesare filtered

        Keyword Arguments:
            underlying_instrument_name {str} -- The reference to the underlying_instrument pk of the outer query (default: {"pk"})

        Returns:
            django.db.models.Subquery -- Subquery containing the sum of shares of each underlying_instrument
        """

        qs = cls.valid_customer_trade_objects
        if not only_customer_trade:
            qs = cls.objects
        qs = qs.filter(
            underlying_instrument=OuterRef(underlying_instrument_name),
            transaction_date__lt=val_date,
        )
        return Coalesce(
            Subquery(
                qs.values("underlying_instrument").annotate(sum_shares=Sum("shares")).values("sum_shares")[:1],
                output_field=models.DecimalField(),
            ),
            Decimal(0),
        )

    def link_to_internal_trade(self):
        qs = Trade.objects.filter(
            Q(underlying_instrument__instrument_type__key="product")
            & Q(shares=self.shares)
            & Q(underlying_instrument=self.underlying_instrument)
            & Q(transaction_date__gte=self.transaction_date - timedelta(days=self.TRADE_WINDOW_INTERVAL))
            & Q(transaction_date__lte=self.transaction_date + timedelta(days=self.TRADE_WINDOW_INTERVAL))
        ).exclude(id=self.id)
        if self.transaction_subtype in [Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION]:
            qs = qs.exclude(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])
            if qs.count() == 1:
                self.internal_trade = qs.first()
                self.save()
        else:
            qs = qs.filter(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])
            if qs.count() == 1:
                trade = qs.first()
                trade.internal_trade = self
                trade.save()

    @classmethod
    def subquery_net_money(
        cls, date_gte=None, date_lte=None, underlying_instrument_name="pk", only_positive=False, only_negative=False
    ):
        """Return a subquery which computes the net negative/positive money per underlying_instrument

        Arguments:
            val_date1 {datetime.date} -- The start date, including
            val_date2 {datetime.date} -- The end date, including

        Keyword Arguments:
            underlying_instrument_name {str} -- The reference to the underlying_instrument pk from the outer query (default: {"pk"})

        Returns:
            django.db.models.Subquery -- The subquery containing the net negative money per underlying_instrument
        """
        qs = cls.valid_external_customer_trade_objects.annotate(
            date_considered=ExpressionWrapper(F("transaction_date") + 1, output_field=DateField())
        )

        if date_gte:
            qs = qs.filter(date_considered__gte=date_gte)
        if date_lte:
            qs = qs.filter(date_considered__lte=date_lte)

        if only_positive:
            qs = qs.filter(shares__gt=0)
        elif only_negative:
            qs = qs.filter(shares__lt=0)
        return Coalesce(
            Subquery(
                qs.filter(underlying_instrument=OuterRef(underlying_instrument_name))
                .annotate(
                    _price=Case(
                        When(
                            price__isnull=True,
                            then=InstrumentPrice.subquery_closest_value(
                                "net_value",
                                date_name="date_considered",
                                instrument_pk_name="underlying_instrument__pk",
                            ),
                        ),
                        default=F("price"),
                    ),
                    net_value=ExpressionWrapper(F("shares") * F("_price"), output_field=models.FloatField()),
                )
                .values("underlying_instrument")
                .annotate(sum_net_value=Sum(F("net_value")))
                .values("sum_net_value"),
                output_field=models.FloatField(),
            ),
            0.0,
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:trade"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:traderepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{|:-}}{{transaction_date}}{{|::}}{{bank}}{{|-:}} {{claimed_shares}} / {{shares}} (∆ {{diff_shares}})"

    @classmethod
    def get_representation_value_key(cls):
        return "id"


@shared_task
def align_custodian():
    unaligned_qs = Trade.objects.annotate(
        proper_custodian_id=Subquery(Custodian.objects.filter(mapping__contains=OuterRef("bank")).values("id")[:1])
    ).exclude(custodian__id=F("proper_custodian_id"))

    unaligned_qs.update(custodian__id=F("proper_custodian_id"))


@receiver(post_save, sender="wbportfolio.Claim")
def compute_claimed_shares_on_claim_save(sender, instance, created, raw, **kwargs):
    if not raw and instance.trade:
        instance.trade.save()


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object: "Instrument", main_object: "Instrument", **kwargs):
    """
    Simply reassign the transactions linked to the merged instrument to the main instrument
    """
    merged_object.trades.update(underlying_instrument=main_object)


@receiver(add_instrument_to_investable_universe, sender="wbfdm.Instrument")
def add_instrument_to_investable_universe_from_transactions(sender: models.Model, **kwargs) -> list[int]:
    """
    register all instrument linked to assets as within the investible universe
    """
    return list(
        (
            Instrument.objects.annotate(
                transaction_exists=models.Exists(Trade.objects.filter(underlying_instrument=models.OuterRef("pk")))
            ).filter(transaction_exists=True)
        )
        .distinct()
        .values_list("id", flat=True)
    )



from datetime import date, timedelta
from datetime import date as date_lib
from decimal import Decimal

from django.db import models
from django.db.models import (
    DateField,
    Exists,
    ExpressionWrapper,
    OuterRef,
    Q,
    QuerySet,
    Sum,
)
from django.db.models.functions import Greatest
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from wbcore.contrib.ai.llm.config import add_llm_prompt
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.signals.models import pre_collection
from wbcore.utils.enum import ChoiceEnum
from wbcore.utils.strings import ReferenceIDMixin
from wbcrm.models import AccountRole
from wbcrm.models.accounts import Account
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

from ..custodians import Custodian
from ..llm.wbcrm.analyze_relationship import get_performances_prompt
from ..roles import PortfolioRole
from .trades import Trade


def can_administrate_claim(claim, user):
    today = date_lib.today()
    return (
        user.profile.portfolio_roles.filter(
            Q(role_type=PortfolioRole.RoleType.MANAGER)
            & (Q(start__isnull=True) | Q(start__lte=today))
            & (Q(end__isnull=True) | Q(end__gte=today))
        ).exists()
        or user.is_superuser
    )


def can_edit_claim(claim, user):
    return (claim.claimant and claim.claimant.id == user.profile.id) or user.profile.is_internal or user.is_superuser


class ClaimGroupbyChoice(ChoiceEnum):
    ROOT_ACCOUNT = "Root Account"
    ACCOUNT = "Account"
    ROOT_ACCOUNT_OWNER = "Root Account Owner"
    ACCOUNT_OWNER = "Account Owner"
    PRODUCT = "Product"
    PRODUCT_GROUP = "ProductGroup"
    CLASSIFICATION = "Classification"

    @classmethod
    @property
    def map(cls):
        return {
            "ROOT_ACCOUNT": {
                "pk": "root_account",
                "title_key": "root_account_repr",
                "search_fields": ["root_account_repr"],
            },
            "ACCOUNT": {
                "pk": "account",
                "title_key": "account__computed_str",
                "search_fields": ["account__computed_str"],
            },
            "ROOT_ACCOUNT_OWNER": {
                "pk": "root_account_owner",
                "title_key": "root_account_owner_repr",
                "search_fields": ["root_account_owner_repr"],
            },
            "ACCOUNT_OWNER": {
                "pk": "account__owner",
                "title_key": "account__owner__computed_str",
                "search_fields": ["account__owner__computed_str"],
            },
            "PRODUCT": {
                "pk": "product",
                "title_key": "product__computed_str",
                "search_fields": ["product__computed_str"],
            },
            "PRODUCT_GROUP": {
                "pk": "product__parent",
                "title_key": "product__parent__name",
                "search_fields": ["product__parent__name"],
            },
            "CLASSIFICATION": {
                "pk": "classification_id",
                "title_key": "classification_title",
                "search_fields": ["classification_title"],
            },
        }

    @classmethod
    def get_map(cls, name):
        return cls.map[name]


class ClaimDefaultQueryset(QuerySet):
    def filter_for_user(self, user: User, validity_date: date_lib | None = None) -> QuerySet:
        """
        Protect the chained queryset and filter the claims that this user cannot see based on the following rules:
        """
        if user.has_perm("wbcrm.administrate_account"):
            return self
        allowed_accounts = Account.objects.filter_for_user(user, validity_date=validity_date).values(
            "id"
        )  # This is way faster
        # accounts = self.annotate(can_see_account=Exists(allowed_accounts.filter(id=OuterRef("account"))))
        if user.profile.is_internal:
            return self.filter(Q(account__isnull=True) | Q(account__in=allowed_accounts))
        return self.filter(Q(account__in=allowed_accounts) | (Q(account__isnull=True) & Q(creator_id=user.profile.id)))

    def filter_for_customer(self, customer: Entry, include_related_roles: bool = False) -> QuerySet:
        """
        Filter the chained queryset to return only the claim that belongs to a certain customer
        """
        customer_account_ids = Account.get_accounts_for_customer(customer).values("id")

        if not include_related_roles:
            return self.filter(account__in=customer_account_ids)
        return self.annotate(
            role_exists=Exists(AccountRole.objects.filter(entry=customer, account=OuterRef("account")))
        ).filter(Q(account__in=customer_account_ids) | Q(role_exists=True))

    def annotate_asset_under_management_for_date(self, val_date: date):
        return self.annotate(
            net_value=InstrumentPrice.subquery_closest_value("net_value", val_date, instrument_pk_name="product"),
            fx_rate=CurrencyFXRates.get_fx_rates_subquery(val_date),
            asset_under_management=models.ExpressionWrapper(
                models.F("net_value") * models.F("shares"),
                output_field=models.DecimalField(max_digits=16, decimal_places=4, default=0.0),
            ),
            asset_under_management_usd=models.ExpressionWrapper(
                models.F("asset_under_management") * models.F("fx_rate"),
                output_field=models.DecimalField(max_digits=16, decimal_places=4, default=0.0),
            ),
        )


class ClaimManager(models.Manager):
    def get_queryset(self) -> ClaimDefaultQueryset:
        return ClaimDefaultQueryset(self.model)

    def filter_for_user(self, user: User, validity_date: date_lib | None = None) -> QuerySet:
        return self.get_queryset().filter_for_user(user, validity_date=validity_date)

    def filter_for_customer(self, customer: Entry, include_related_roles: bool = False) -> QuerySet:
        return self.get_queryset().filter_for_customer(customer)

    def annotate_asset_under_management_for_date(self, val_date: date) -> QuerySet:
        return self.get_queryset().annotate_asset_under_management_for_date(val_date)


# @workflow.register(serializer_class="wbportfolio.serializers.transactions.claim.ClaimModelSerializer") #we don't register for now. Uncomment as soon as we want to enable workflow on that model
class Claim(ReferenceIDMixin, WBModel):
    """A customer can claim that a trade or part of a trade was executed my them"""

    class Status(models.TextChoices):
        PENDING = "PENDING", "Pending"
        APPROVED = "APPROVED", "Approved"
        WITHDRAWN = "WITHDRAWN", "Withdrawn"
        DRAFT = "DRAFT", "Draft"
        AUTO_MATCHED = "AUTO_MATCHED", "Auto-Matched"

    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    account = models.ForeignKey(
        to="wbcrm.Account",
        related_name="claims",
        null=True,
        blank=True,
        limit_choices_to=models.Q(is_terminal_account=True),
        on_delete=models.SET_NULL,
        verbose_name="Account",
        # help_text="The account the claim is assigned to. If no sub-account is provided it will be tried to be assigned to a sub account based on the given claimant."
    )

    trade = models.ForeignKey(
        to="wbportfolio.Trade",
        related_name="claims",
        blank=True,
        null=True,
        on_delete=models.PROTECT,  # We protect the claim in case of trade deletion. This needs to be handled upstream
        verbose_name="Trade",
        help_text="Please select a Product first. The customer-trade the claim is consolidated against. The customer-trade and the claim don't necessarily have to have the same date, number of shares, etc.",
    )
    product = models.ForeignKey(
        to="wbportfolio.Product", related_name="claims", on_delete=models.PROTECT, verbose_name="Product"
    )
    claimant = models.ForeignKey(
        to="directory.Entry",
        related_name="claimed",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Claimant",
        # help_text="The person / company that claims this trade. If no claimant is given, it will assigned to the current user."
    )
    creator = models.ForeignKey(
        to="directory.Entry",
        related_name="claims_created",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Creator",
    )
    date = models.DateField(verbose_name="Trade Date")
    bank = models.CharField(max_length=255, blank=True, verbose_name="Bank")
    reference = models.CharField(max_length=255, null=True, blank=True, verbose_name="Additional Reference")
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        help_text="The amount of shares purchased / sold.",
        null=True,
        blank=True,
        verbose_name="Shares",
    )

    nominal_amount = models.DecimalField(
        max_digits=15,
        decimal_places=2,
        help_text="The nominal amount purchased / sold. Either shares or nominal_amount has to be provided.",
        null=True,
        blank=True,
        verbose_name="Nominal Amount",
    )
    external_id = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="An External ID to reference a claim.",
        verbose_name="External Identifier",
    )
    as_shares = models.BooleanField(default=True)

    objects = ClaimManager()

    class Meta:
        verbose_name = "Claim"
        verbose_name_plural = "Claims"

    def __str__(self):
        return f"{self.reference_id} {self.product.name} ({self.bank} - {self.shares:,} shares - {self.date}) "

    def save(self, *args, auto_match: bool = True, **kwargs):
        if self.shares is None and self.nominal_amount is None:
            raise ValueError(
                f"Either shares or nominal amount have to be provided. Shares={self.shares}, Nominal={self.nominal_amount}"
            )
        if self.product:
            if self.shares is not None:
                self.nominal_amount = self.shares * self.product.share_price
            else:
                self.shares = self.nominal_amount / self.product.share_price
        if not self.trade and self.status == self.Status.DRAFT and auto_match:
            self.trade = self.auto_match()
        if self.status == self.Status.WITHDRAWN and self.trade:
            self.trade = None
        if self.trade:
            if not self.trade.is_claimable:
                raise ValueError("The selected trade is not a valid customer trade")
            if self.product and self.trade.underlying_instrument.id != self.product.id:
                raise ValueError("The selected product does not match the trade underlying instrument")
        super().save(*args, **kwargs)

    @classmethod
    def get_valid_and_approved_claims(cls, val_date: date_lib | None = None, account: Account | None = None):
        claims = cls.objects.annotate(
            date_considered=ExpressionWrapper(
                Greatest("trade__transaction_date", "date") + 1, output_field=DateField()
            )
        ).filter(Q(account__is_terminal_account=True) & Q(account__is_active=True) & Q(status=cls.Status.APPROVED))
        if val_date:
            claims = claims.filter(date_considered__lt=val_date)
        if account:
            claims = claims.filter(account__in=account.get_descendants(include_self=True))
        return claims

    @property
    def nominal_value(self):
        """Returns the nominal value of a claim

        Shares x Share Price

        Returns:
            Decimal -- Nominal Value
        """
        return self.shares * self.product.share_price

    @transition(
        status,
        [Status.PENDING, Status.AUTO_MATCHED],
        Status.APPROVED,
        permission=can_administrate_claim,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.APPROVE.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="approve",
                label="Approve",
                action_label="Approve",
                description_fields="<p>Date: {{date}}</p><p>Shares: {{shares}}</p><p>Product: {{_product.name}} ({{_product.isin}})</p>",
                instance_display=create_simple_display(
                    [
                        ["date", "shares", "bank"],
                        ["product", "claimant", "."],
                        ["account", "account", "account"],
                        ["trade", "trade", "trade"],
                    ]
                ),
            )
        },
    )
    def approve(self, **kwargs):
        pass

    def can_approve(self):
        errors = dict()

        if not self.trade:
            errors["trade"] = [_("With this status, this has to be provided.")]

        if not self.product:
            errors["product"] = [_("With this status, this has to be provided.")]

        if not self.account:
            errors["account"] = [_("With this status, this has to be provided.")]

        # check if the specified product have a valid nav at the specified date
        if (
            (product := self.product)
            and (claim_date := self.date)
            and not product.valuations.filter(date=claim_date).exists()
        ):
            if (prices_qs := product.valuations.filter(date__lt=claim_date)).exists():
                errors["date"] = [
                    f"For product {product.name}, the latest valid valuation date before {claim_date:%Y-%m-%d} is {prices_qs.latest('date').date:%Y-%m-%d}: Please select a valid date."
                ]
            else:
                errors["date"] = [
                    f"There is no valuation before {claim_date:%Y-%m-%d} for product {product.name}: Please select a valid date."
                ]
        return errors

    @transition(
        status,
        [Status.PENDING, Status.AUTO_MATCHED],
        Status.DRAFT,
        permission=lambda claim, user: user.profile.is_internal or user.is_superuser,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.ERROR,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="Back to Draft",
                description_fields="<p>Date: {{date}}</p><p>Shares: {{shares}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        pass

    @transition(
        status,
        [Status.DRAFT],
        Status.WITHDRAWN,
        permission=can_edit_claim,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                color=ButtonDefaultColor.ERROR,
                icon=WBIcon.DELETE.icon,
                key="withdraw",
                label="Withdraw",
                action_label="Withdraw",
                description_fields="<p>Date: {{date}}</p><p>Shares: {{shares}}</p>",
            )
        },
    )
    def withdraw(self, **kwargs):
        pass

    @transition(
        status,
        [Status.APPROVED],
        Status.DRAFT,
        permission=lambda claim, user: user.profile.is_internal or user.is_superuser,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.EDIT.icon,
                color=ButtonDefaultColor.WARNING,
                key="revise",
                label="Revise",
                action_label="Revise",
                description_fields="<p>Date: {{date}}</p><p>Product: {{_product.name}}</p>",
            )
        },
    )
    def revise(self, **kwargs):
        pass

    @transition(
        status,
        [Status.DRAFT, Status.AUTO_MATCHED],
        Status.PENDING,
        permission=lambda claim, user: user.profile.is_internal or user.is_superuser,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.SEND.icon,
                color=ButtonDefaultColor.WARNING,
                key="submit",
                label="Submit for approval",
                action_label="Submit",
                description_fields="<p>Date: {{date}}</p><p>Product: {{_product.name}}</p>",
            )
        },
    )
    def submit(self, **kwargs):
        pass

    def can_submit(self):
        return self.can_approve()

    def auto_match(self) -> Trade | None:
        shares_epsilon = 1  # share
        auto_match_trade = None
        # Obvious filtering
        trades = Trade.valid_customer_trade_objects.filter(
            Q(transaction_date__gte=self.date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL))
            & Q(transaction_date__lte=self.date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL))
        )
        if self.product:
            trades = trades.filter(underlying_instrument=self.product)
        # Find trades by shares (or remaining to be claimed)
        trades = trades.filter(
            Q(diff_shares__lte=self.shares + shares_epsilon) & Q(diff_shares__gte=self.shares - shares_epsilon)
        )
        if trades.count() == 1:
            auto_match_trade = trades.first()

        # Find trades by custodian
        if (
            not auto_match_trade
            and self.bank
            and (custodian := Custodian.get_by_mapping(self.bank, use_similarity=True, create_missing=False))
        ):
            if trades.filter(custodian=custodian).exists():
                trades = trades.filter(custodian=custodian)
            if trades.count() == 1:
                auto_match_trade = trades.first()

        # Find trades by external_id
        if not auto_match_trade and self.external_id and trades.count() > 1:
            trades = trades.filter(
                Q(external_id__icontains=self.external_id) | Q(external_id_alternative__icontains=self.external_id)
            )
            if trades.count() == 1:
                auto_match_trade = trades.first()

        if auto_match_trade:
            self.status = self.Status.AUTO_MATCHED
            return auto_match_trade

    @classmethod
    def subquery_assets_under_management_for_account(cls, claims, price_date, account_key="account__pk"):
        """Returns a subquery which annotates the assets in USD for each Sub Account based on a given queryset

        Arguments:
            claims {QuerySet<commission.Claim>} -- Prefiltered queryset of claims
            price_date {datetime.date} -- the date which is used to calculate everything
            account_key {str} -- the outer reference to the sub account pk (default: account__pk)

        Returns:
            Subquery -- Sub Account with assets in USD
        """

        return models.Subquery(
            claims.filter(
                status=cls.Status.APPROVED,
                account__pk=models.OuterRef(account_key),
                trade__transaction_date__lte=price_date,
            )
            .annotate_asset_under_management_for_date(price_date)
            .values("account")
            .annotate(sum_assets_usd=models.Sum("assets_under_management_usd"))
            .values("sum_assets_usd")[:1],
            output_field=models.DecimalField(max_digits=16, decimal_places=4, default=0.0),
        )

    @classmethod
    def subquery_claim_sum_for_account(cls, claims, price_date, account_key="account__pk"):
        return models.Subquery(
            claims.filter(
                status=cls.Status.APPROVED,
                account__pk=models.OuterRef(account_key),
                trade__transaction_date__lte=price_date,
            )
            .values("account")
            .annotate(claim_sum=models.Sum("shares"))
            .values("claim_sum")[:1],
            output_field=models.DecimalField(max_digits=16, decimal_places=0, default=0.0),
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:claim"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:claimrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{shares}} ({{account}})"


@receiver(models.signals.post_save, sender="wbportfolio.Trade")
def post_save_trade(sender, instance, created, raw, **kwargs):
    """
    For every claimable trade, first try to autoclaim the already existing draft claims with this new trade. And auto claim them as well if the product has a default sub account
    """
    if not raw and instance.is_claimable and created:
        min_transaction_date = instance.transaction_date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL)
        max_transaction_date = instance.transaction_date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL)
        unlinked_claims = Claim.objects.filter(
            status=Claim.Status.DRAFT,
            date__gte=min_transaction_date,
            date__lte=max_transaction_date,
            trade__isnull=True,
            product__id=instance.underlying_instrument.id,
        )
        for claim in unlinked_claims.all():
            claim.trade = claim.auto_match()
            claim.save()
        if instance.product and (account := instance.product.default_account):
            shares = instance.shares or Decimal(0)
            claimed_shares = instance.claims.filter(status=Claim.Status.APPROVED).aggregate(s=models.Sum("shares"))[
                "s"
            ] or Decimal(0)
            rest_shares = shares - claimed_shares
            if rest_shares != 0:
                Claim.objects.create(
                    trade=instance,
                    shares=rest_shares,
                    bank=instance.bank,
                    date=instance.transaction_date,
                    product=instance.product,
                    account=account,
                    claimant=account.owner,
                )
            for claim in instance.claims.exclude(status=Claim.Status.APPROVED):
                if claim.status == Claim.Status.DRAFT:
                    claim.submit()
                if claim.status == Claim.Status.PENDING:
                    claim.approve()
                claim.save()


@receiver(pre_collection, sender="wbportfolio.Trade")
def pre_trade_deletion(sender, instance, **kwargs):
    # We can't use pre_delete signal here because the collector is collected before the signal and thus, it will
    for claim in instance.claims.filter(
        status__in=[Claim.Status.DRAFT, Claim.Status.AUTO_MATCHED, Claim.Status.WITHDRAWN]
    ).all():
        claim.trade = None
        claim.save(auto_match=False)

    if (instance.marked_for_deletion or instance.pending) and instance.claims.exists():
        # If a default account exists on the trade's product, it means that the product's trade are automatically claims. In that case, we are sure that a valid trade still exists and the marked for deletion's claim can be safely remove
        if instance.product and instance.product.default_account:
            instance.claims.all().delete()
        else:
            similar_trades = Trade.objects.filter(
                underlying_instrument=instance.underlying_instrument,
                portfolio=instance.portfolio,
                marked_for_deletion=True,
                transaction_date=instance.transaction_date,
            )
            # We check if the sum of marked for deletion trades share sums to 0, in that case, we delete them without regards and set their potential claims to draft
            if similar_trades.exists() and similar_trades.aggregate(s=Sum("shares"))["s"] == 0:
                for t in similar_trades.all():
                    t.claims.update(trade=None, status=Claim.Status.DRAFT)
            elif (other_unclaims_similar_trades := instance.get_alternative_valid_trades()).exists():
                if other_unclaims_similar_trades.count() > 1:
                    other_unclaims_similar_trades = other_unclaims_similar_trades.filter(bank=instance.bank)
                if other_trade := other_unclaims_similar_trades.first():
                    # If we find an unclaim trade with similar attributes, we forward the marked_for_deletion attribute to it, which will be handled/deleted in a next delete iteration
                    instance.claims.update(trade=other_trade)


@receiver(pre_merge, sender="wbcrm.Account")
def handle_pre_merge_account_for_claim(
    sender: models.Model, merged_object: "Account", main_object: "Account", **kwargs
):
    """
    Simply reassign the claim linked to the merged account to the main account
    """
    Claim.objects.filter(account=merged_object).update(account=main_object, reference=merged_object.reference_id)


@receiver(add_llm_prompt, sender="wbcrm.Account")
def add_holdings_to_account_heat(sender, instance, key, **kwargs):
    if key == "analyze_relationship":
        return get_performances_prompt(instance)
    return []



from typing import Generator

from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcompliance.models.risk_management.rules import RiskIncidentType
from wbcore import serializers as wb_serializers
from wbfdm.enums import ESGControveryFlag
from wbfdm.models import Instrument
from wbfdm.models.esg.controversies import Controversy

from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .mixins import ActivePortfolioRelationshipMixin


@register("Controversy Portfolio Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActivePortfolioRelationshipMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # if thresholds are attached to this rule, we take the first as severity. Otherwise, we get the risk incident with the highest severity (e.g. Critical)
        if self.thresholds:
            self.severity = self.thresholds[0].severity
        else:
            self.severity = RiskIncidentType.objects.order_by("-severity_order").first()
        self.flags_repr = [ESGControveryFlag[f].label for f in self.flags]

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            flags = wb_serializers.MultipleChoiceField(
                choices=ESGControveryFlag.choices,
                default=[ESGControveryFlag.ORANGE.value, ESGControveryFlag.RED.value],
                label="Flags",
                help_text="Set the flags that will trigger the rule",
            )

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "flags",
                ]

        return RuleBackendSerializer

    def _process_dto(self, portfolio: PortfolioDTO, **kwargs) -> Generator[backend.IncidentResult, None, None]:
        for instrument_id in portfolio.positions_map.keys():
            instrument = Instrument.objects.get(id=instrument_id)
            if (
                controversies := Controversy.objects.filter(
                    instrument__in=instrument.get_ancestors(include_self=True), flag__in=self.flags
                )
            ).exists():
                controversies_headlines = "".join([f"<li>{c.headline}</li>" for c in controversies])
                yield backend.IncidentResult(
                    breached_object=instrument,
                    breached_object_repr=str(instrument),
                    breached_value=f"# {controversies.count()}",
                    report_details={
                        "Controversies Flags": ", ".join(self.flags_repr),
                        "Controversies Headlines": controversies_headlines,
                    },
                    severity=self.severity,
                )



from decimal import Decimal

from rest_framework import serializers
from wbcore import serializers as wb_serializers
from wbcore.metadata.configs.display.list_display import BaseTreeGroupLevelOption
from wbfdm.models import Instrument
from wbfdm.serializers import InvestableInstrumentRepresentationSerializer
from wbfdm.serializers.instruments.instruments import (
    CompanyRepresentationSerializer,
    SecurityRepresentationSerializer,
)

from wbportfolio.models import Order


class GetSecurityDefault:
    requires_context = True

    def __call__(self, serializer_instance):
        try:
            instance = serializer_instance.view.get_object()
            return instance.underlying_instrument.parent or instance.underlying_instrument
        except Exception:
            return None


class GetCompanyDefault:
    requires_context = True

    def __call__(self, serializer_instance):
        try:
            instance = serializer_instance.view.get_object()
            security = instance.underlying_instrument.parent or instance.underlying_instrument
            return security.parent or security
        except Exception:
            return None


class OrderOrderProposalListModelSerializer(wb_serializers.ModelSerializer):
    underlying_instrument = wb_serializers.SlugRelatedField(read_only=True, slug_field="name")
    underlying_instrument_isin = wb_serializers.CharField(read_only=True)
    underlying_instrument_ticker = wb_serializers.CharField(read_only=True)
    underlying_instrument_refinitiv_identifier_code = wb_serializers.CharField(read_only=True)
    underlying_instrument_instrument_type = wb_serializers.CharField(read_only=True)
    underlying_instrument_exchange = wb_serializers.CharField(read_only=True)

    effective_weight = wb_serializers.DecimalField(
        read_only=True,
        max_digits=Order.ORDER_WEIGHTING_PRECISION + 1,
        decimal_places=Order.ORDER_WEIGHTING_PRECISION,
        default=0,
    )
    target_weight = wb_serializers.DecimalField(
        max_digits=Order.ORDER_WEIGHTING_PRECISION + 1,
        decimal_places=Order.ORDER_WEIGHTING_PRECISION,
        required=False,
    )

    effective_shares = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)
    target_shares = wb_serializers.DecimalField(required=False, max_digits=16, decimal_places=6)

    effective_total_value_fx_portfolio = wb_serializers.DecimalField(
        read_only=True, max_digits=16, decimal_places=2, default=0
    )
    target_total_value_fx_portfolio = wb_serializers.DecimalField(required=False, max_digits=16, decimal_places=2)
    total_value_fx_portfolio = wb_serializers.DecimalField(required=False, max_digits=16, decimal_places=2)

    portfolio_currency = wb_serializers.CharField(read_only=True)
    has_warnings = wb_serializers.BooleanField(read_only=True)

    def validate(self, data):
        data.pop("company", None)
        data.pop("security", None)
        if self.instance and "underlying_instrument" in data:
            raise serializers.ValidationError(
                {
                    "underlying_instrument": "You cannot modify the underlying instrument other than creating a new entry"
                }
            )
        effective_weight = self.instance._effective_weight if self.instance else Decimal(0.0)
        effective_shares = self.instance._effective_shares if self.instance else Decimal(0.0)
        portfolio_value = (
            self.context["view"].order_proposal.portfolio_total_asset_value if "view" in self.context else Decimal(0.0)
        )
        if (total_value_fx_portfolio := data.pop("total_value_fx_portfolio", None)) is not None and portfolio_value:
            data["weighting"] = total_value_fx_portfolio / portfolio_value
        if (
            target_total_value_fx_portfolio := data.pop("target_total_value_fx_portfolio", None)
        ) is not None and portfolio_value:
            data["target_weight"] = target_total_value_fx_portfolio / portfolio_value

        if data.get("weighting") or data.get("target_weight"):
            weighting = data.pop("weighting", None)
            if (target_weight := data.pop("target_weight", None)) is not None:
                weighting = target_weight - effective_weight
                data["desired_target_weight"] = target_weight
            if weighting is not None:
                data["weighting"] = weighting
            data.pop("shares", None)
            data.pop("target_shares", None)

        if data.get("shares") or data.get("target_shares"):
            shares = data.pop("shares", None)
            if (target_shares := data.pop("target_shares", None)) is not None:
                shares = target_shares - effective_shares
            if shares is not None:
                data["shares"] = shares
        return super().validate(data)

    def update(self, instance, validated_data):
        weighting = validated_data.pop("weighting", None)
        shares = validated_data.pop("shares", None)
        portfolio_total_asset_value = instance.order_proposal.portfolio_total_asset_value
        if weighting is not None:
            instance.set_weighting(weighting, portfolio_total_asset_value)
        if shares is not None:
            instance.set_shares(shares, portfolio_total_asset_value)
        return super().update(instance, validated_data)

    def create(self, validated_data):
        weighting = validated_data.pop("weighting", None)
        shares = validated_data.pop("shares", None)
        instance = super().create(validated_data)
        portfolio_total_asset_value = instance.order_proposal.portfolio_total_asset_value
        if weighting is not None:
            instance.set_weighting(weighting, portfolio_total_asset_value)
        if shares is not None:
            instance.set_shares(shares, portfolio_total_asset_value)
        instance.save()
        return instance

    class Meta:
        model = Order
        percent_fields = ["effective_weight", "target_weight", "weighting"]
        decorators = {
            "total_value_fx_portfolio": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{portfolio_currency}}"
            ),
            "effective_total_value_fx_portfolio": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{portfolio_currency}}"
            ),
            "target_total_value_fx_portfolio": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{portfolio_currency}}"
            ),
        }
        read_only_fields = (
            "order_type",
            "effective_shares",
            "effective_total_value_fx_portfolio",
            "has_warnings",
        )
        fields = (
            "id",
            "shares",
            "underlying_instrument",
            "underlying_instrument_isin",
            "underlying_instrument_ticker",
            "underlying_instrument_refinitiv_identifier_code",
            "underlying_instrument_instrument_type",
            "underlying_instrument_exchange",
            "order_type",
            "comment",
            "effective_weight",
            "target_weight",
            "weighting",
            "order_proposal",
            "order",
            "effective_shares",
            "target_shares",
            "total_value_fx_portfolio",
            "effective_total_value_fx_portfolio",
            "target_total_value_fx_portfolio",
            "portfolio_currency",
            "has_warnings",
            "execution_confirmed",
            "execution_comment",
        )


class OrderOrderProposalModelSerializer(OrderOrderProposalListModelSerializer):
    company = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.filter(level=0),
        required=False,
        read_only=lambda view: not view.new_mode,
        default=GetCompanyDefault(),
    )
    _company = CompanyRepresentationSerializer(source="company", required=False)

    security = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.filter(is_security=True),
        required=False,
        read_only=lambda view: not view.new_mode,
        default=GetSecurityDefault(),
    )
    _security = SecurityRepresentationSerializer(
        source="security",
        optional_get_parameters={"company": "parent"},
        depends_on=[{"field": "company", "options": {}}],
        required=False,
    )
    underlying_instrument = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.all(), label="Quote", read_only=lambda view: not view.new_mode
    )
    _underlying_instrument = InvestableInstrumentRepresentationSerializer(
        source="underlying_instrument",
        optional_get_parameters={"security": "parent"},
        depends_on=[{"field": "security", "options": {}}],
        tree_config=BaseTreeGroupLevelOption(clear_filter=True, filter_key="parent"),
    )

    class Meta(OrderOrderProposalListModelSerializer.Meta):
        fields = list(OrderOrderProposalListModelSerializer.Meta.fields) + [
            "company",
            "_company",
            "security",
            "_security",
            "_underlying_instrument",
        ]


class ReadOnlyOrderOrderProposalModelSerializer(OrderOrderProposalListModelSerializer):
    class Meta(OrderOrderProposalListModelSerializer.Meta):
        read_only_fields = OrderOrderProposalListModelSerializer.Meta.fields



from contextlib import suppress

from django.db.utils import ProgrammingError
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.directory.models import CustomerStatus
from wbcore.contrib.dynamic_preferences.types import (
    CallableDefaultModelChoicePreference,
)

portfolio = Section("wbportfolio")


def get_default_returning_client_customer_status():
    with suppress(ProgrammingError):
        return CustomerStatus.objects.filter(title="Returning Client").first()


def get_default_lost_client_customer_status():
    with suppress(ProgrammingError):
        return CustomerStatus.objects.filter(title="Lost Client").first()


def get_default_tpm_customer_status():
    with suppress(ProgrammingError):
        return CustomerStatus.objects.filter(title="Third Party Marketer (TPM)").first()


def get_default_client_customer_status():
    with suppress(ProgrammingError):
        return CustomerStatus.objects.filter(title="Client").first()


@global_preferences_registry.register
class ReturningClientCustomerStatus(CallableDefaultModelChoicePreference):
    section = portfolio
    name = "returning_client_customer_status"
    model = CustomerStatus

    verbose_name = "Returning Client customer status"
    help_text = "The customer status corresponding to the Returning Client Group"

    @property
    def default(self):
        return get_default_returning_client_customer_status()


@global_preferences_registry.register
class LostClientCustomerStatus(CallableDefaultModelChoicePreference):
    section = portfolio
    name = "lost_client_customer_status"
    model = CustomerStatus

    verbose_name = "Lost Client customer status"
    help_text = "The customer status corresponding to the Lost Client Group"

    @property
    def default(self):
        return get_default_lost_client_customer_status()


@global_preferences_registry.register
class TPMCustomerStatus(CallableDefaultModelChoicePreference):
    section = portfolio
    name = "tpm_customer_status"
    model = CustomerStatus

    verbose_name = "TPM customer status"
    help_text = "The customer status corresponding to the TPM group"

    @property
    def default(self):
        return get_default_tpm_customer_status()


@global_preferences_registry.register
class ClientCustomerStatus(CallableDefaultModelChoicePreference):
    section = portfolio
    name = "client_customer_status"
    model = CustomerStatus

    verbose_name = "Client customer status"
    help_text = "The customer status corresponding to the Client group"

    @property
    def default(self):
        return get_default_client_customer_status()



from contextlib import suppress

from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbportfolio.models import OrderProposal


class OrderOrderProposalEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if order_proposal_id := self.view.kwargs.get("order_proposal_id", None):
            order_proposal = OrderProposal.objects.get(id=order_proposal_id)
            if order_proposal.status == OrderProposal.Status.DRAFT:
                return reverse(
                    "wbportfolio:orderproposal-order-list",
                    args=[self.view.kwargs["order_proposal_id"]],
                    request=self.request,
                )
        return None

    def get_delete_endpoint(self, **kwargs):
        with suppress(AttributeError, AssertionError):
            order = self.view.get_object()
            if order._effective_weight:  # we make sure order with a valid effective position cannot be deleted
                return None
        return super().get_delete_endpoint(**kwargs)



from django.db.models import Q
from django.utils import timezone
from wbcore import filters as wb_filters

from wbportfolio.models import PortfolioRole


class PortfolioRoleFilterSet(wb_filters.FilterSet):
    is_active = wb_filters.BooleanFilter(method="filter_is_active", initial=True, label="Is Active")

    def filter_is_active(self, queryset, name, value):
        if value is True:
            return queryset.filter(Q(end__isnull=True) | Q(end__gte=timezone.now())).distinct()
        elif value is False:
            return queryset.filter(end__lte=timezone.now())
        return queryset

    class Meta:
        model = PortfolioRole
        fields = {
            "role_type": ["exact"],
            "person": ["exact"],
            "start": ["gte", "exact", "lte"],
            "end": ["gte", "exact", "lte"],
            "weighting": ["gte", "exact", "lte"],
            "instrument": ["exact"],
        }



from zipfile import BadZipFile

import numpy as np
import pandas as pd

from .utils import _get_underlying_instrument, file_name_parse_isin

FIELD_MAP = {
    "Close": "initial_price",
    "Nb of shares": "initial_shares",
    "Fx Rate": "initial_currency_fx_rate",
    "Valuation Date": "asset_valuation_date",
    "Quoted Crncy": "currency__key",
    "exchange": "exchange",
    "underlying_quote": "underlying_quote",
}


def _apply_adjusting_factor(row):
    """
    If the position is a product position, then the adjusting factor adjusts the shares otherwise, it adjusts the price.

    No idea why though
    """
    if row["underlying_quote"]["instrument_type"] == "product":
        return pd.Series([row["initial_price"], row["initial_shares"] * row["Quotity/Adj. factor"]])
    else:
        return pd.Series([row["initial_price"] * row["Quotity/Adj. factor"], row["initial_shares"]])


def parse(import_source):
    # Parse the Parts of the filename into the different parts
    parts = file_name_parse_isin(import_source.file.name)
    # Get the valuation date and investment from the parts list
    valuation_date = parts["valuation_date"]
    product_data = parts["product"]

    # Load file into a CSV DictReader
    if import_source.file.name.lower().endswith(".csv"):
        df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    else:
        try:
            df = pd.read_excel(import_source.file, engine="openpyxl", sheet_name="Basket Valuation")
        except BadZipFile:
            df = pd.read_excel(import_source.file, engine="xlrd", sheet_name="Basket Valuation")
        xx, yy = np.where(df == "Ticker")
        if not xx and not yy:
            xx, yy = np.where(df == "Code")
        df = df.iloc[xx[0] :, yy[0] :]
        df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all")
        df["Quotity/Adj. factor"] = 1.0
        df = df.rename(columns={"Code": "Ticker"})

    df = df.rename(columns=FIELD_MAP)
    df = df.dropna(subset=["initial_price"])
    df["initial_price"] = df["initial_price"].astype("str").str.replace(" ", "").astype("float")
    df["underlying_quote"] = df[["Ticker", "Name", "currency__key"]].apply(
        lambda x: _get_underlying_instrument(*x), axis=1
    )
    df["initial_price"] = df["initial_price"].replace(0, np.nan).fillna(1.0)
    df[["initial_price", "initial_shares"]] = df[
        ["initial_price", "initial_shares", "Quotity/Adj. factor", "underlying_quote"]
    ].apply(lambda x: _apply_adjusting_factor(x), axis=1)
    df["exchange"] = df.underlying_quote.apply(lambda x: x.get("exchange", None))
    df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))

    df["portfolio__instrument_type"] = "product"
    if "isin" in product_data:
        df["portfolio__isin"] = product_data["isin"]
    if "ticker" in product_data:
        df["portfolio__ticker"] = product_data["ticker"]
    df["is_estimated"] = False
    df["date"] = valuation_date.strftime("%Y-%m-%d")
    df["asset_valuation_date"] = pd.to_datetime(df["asset_valuation_date"], dayfirst=True).dt.strftime("%Y-%m-%d")

    df["weighting"] = df.initial_currency_fx_rate * df.initial_price * df.initial_shares
    df["weighting"] = df.weighting / df.weighting.sum()
    return {"data": df.to_dict("records")}



from datetime import timedelta
from typing import Generator

import pandas as pd
from django.contrib.contenttypes.models import ContentType
from django.contrib.humanize.templatetags.humanize import intcomma
from django.db import models
from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.models import Entry
from wbcrm.models import Account
from wbfdm.models import Classification
from wbfdm.preferences import get_default_classification_group

from wbportfolio.analysis.claims import ConsolidatedTradeSummary
from wbportfolio.models import Product, ProductGroup
from wbportfolio.models.transactions.claim import Claim, ClaimGroupbyChoice
from wbportfolio.serializers import ProductRepresentationSerializer


@register("Account Shares Rule Backend", rule_group_key="sales")
class RuleBackend(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "customer"

    customer: Entry

    class FieldChoices(models.TextChoices):
        SHARES = "SHARES", "Shares"
        AUM = "AUM", "AUM"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.end_date = self.evaluation_date
        self.start_date = (self.evaluation_date - pd.tseries.offsets.BDay(self.business_days_interval)).date()

        self.group_by = ClaimGroupbyChoice[self.group_by]
        self.field = self.FieldChoices[self.field]

    def is_passive_evaluation_valid(self) -> bool:
        return Claim.objects.filter_for_customer(self.customer).filter(date__lte=self.evaluation_date).exists()

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Entry)

    def _build_dto_args(self):
        qs = Claim.objects.filter_for_customer(self.customer).filter(
            status=Claim.Status.APPROVED, date__lte=self.evaluation_date
        )
        if self.only_products:
            qs = qs.filter(product__in=self.only_products)
        groupby_map = ClaimGroupbyChoice.get_map(self.group_by.name)
        pivot = groupby_map["pk"]
        pivot_label = groupby_map["title_key"]
        cts_generator = ConsolidatedTradeSummary(
            qs,
            self.start_date,
            self.end_date + timedelta(days=1),  # we shift by one because end date is excluded
            pivot,
            pivot_label,
            classification_group=get_default_classification_group(),
        )
        return (cts_generator,)

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        return Entry.objects.annotate(
            has_open_account=models.Exists(Account.open_objects.filter(owner=models.OuterRef("pk")))
        ).filter(has_open_account=True)

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            business_days_interval = wb_serializers.IntegerField(default=7)
            moving_average_window = wb_serializers.IntegerField(default=1)  # 1 means the initial time series

            only_products = wb_serializers.PrimaryKeyRelatedField(
                queryset=Product.objects.all(),
                many=True,
                default=None,
                allow_null=True,
                label="Only Products",
            )
            _only_products = ProductRepresentationSerializer(many=True, source="parameters__only_products")

            group_by = wb_serializers.ChoiceField(
                choices=ClaimGroupbyChoice.choices(),
                default=ClaimGroupbyChoice.ACCOUNT,
                allow_null=True,
                help_text="Choose how to group by shares",
                label="Group By",
            )
            field = wb_serializers.ChoiceField(
                choices=cls.FieldChoices.choices,
                default=cls.FieldChoices.SHARES,
                allow_null=True,
                help_text="Choose which metric to choose",
                label="Field",
            )

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "field",
                    "group_by",
                    "business_days_interval",
                    "moving_average_window",
                    "only_products",
                ]

        return RuleBackendSerializer

    def _process_dto(
        self, cts_generator: ConsolidatedTradeSummary, **kwargs
    ) -> Generator[backend.IncidentResult, None, None]:
        df = cts_generator.get_aum_df()
        if df.empty:
            return
        if self.field == self.FieldChoices.SHARES:
            perf = df["sum_shares_perf"]
            start_df = df["sum_shares_start"]
            end_df = df["sum_shares_end"]
        else:
            perf = df["sum_aum_perf"]
            start_df = df["sum_aum_start"]
            end_df = df["sum_aum_end"]
        perf = perf.dropna()
        if not perf.empty:
            for threshold in self.thresholds:
                numerical_range = threshold.numerical_range
                breached_perf = perf[(perf > numerical_range[0]) & (perf < numerical_range[1])].dropna()
                if not breached_perf.empty:
                    for breached_obj_id, percentage in breached_perf.to_dict().items():
                        breached_obj = None
                        if self.group_by == ClaimGroupbyChoice.PRODUCT:
                            breached_obj = Product.objects.get(id=breached_obj_id)
                        elif self.group_by == ClaimGroupbyChoice.PRODUCT_GROUP:
                            breached_obj = ProductGroup.objects.get(id=breached_obj_id)
                        elif self.group_by == ClaimGroupbyChoice.CLASSIFICATION:
                            breached_obj = Classification.objects.get(id=breached_obj_id)
                        elif self.group_by in [ClaimGroupbyChoice.ACCOUNT, ClaimGroupbyChoice.ROOT_ACCOUNT]:
                            breached_obj = Account.objects.get(id=breached_obj_id)
                        elif self.group_by in [
                            ClaimGroupbyChoice.ACCOUNT_OWNER,
                            ClaimGroupbyChoice.ROOT_ACCOUNT_OWNER,
                        ]:
                            breached_obj = Entry.objects.get(id=breached_obj_id)
                        report_details = {
                            "Period": f"{cts_generator.start_date:%d.%m.%Y} - {cts_generator.end_date:%d.%m.%Y}",
                        }

                        # create report detail template
                        color = "red" if percentage < 0 else "green"
                        number_prefix = ""

                        if self.field == "AUM":
                            key = "AUM Change"
                            number_prefix = "$"
                        else:
                            key = "Shares Change"

                        template = f'{number_prefix}{{}} → {number_prefix}{{}} <span style="color:{color}"><strong>Δ {number_prefix}{{}}</strong></span>'
                        start = intcomma(int(start_df.loc[breached_obj_id].round(0)))
                        end = intcomma(int(end_df.loc[breached_obj_id].round(0)))
                        diff = intcomma(int((end_df.loc[breached_obj_id] - start_df.loc[breached_obj_id]).round(0)))
                        report_details[key] = template.format(start, end, diff)
                        report_details["Group By"] = self.group_by.value
                        yield backend.IncidentResult(
                            breached_object=breached_obj,
                            breached_object_repr=str(breached_obj),
                            breached_value=f'<span style="color:{color}">{percentage:+,.2%}</span>',
                            report_details=report_details,
                            severity=threshold.severity,
                        )
